import torch
from numpy import ndarray
from torch import Tensor

from .base import BaseOps
from .node import DataNode
from .utils import generate_key


class DatasetOps(BaseOps):
    def add_dataset(self, path: str, name: str, data: Tensor | ndarray | None = None):
        """Adds a single dataset to a specified path in the DataContainer.

        Args:
            path (str): The path to the parent group.
            name (str): The name of the dataset to add.
            data (Tensor, optional): The data to store in the dataset. If None, an empty dataset is created.

        Raises:
            KeyError: If the parent group does not exist.
            KeyError: If the dataset already exists.
        """
        key, _, _ = generate_key(path, name)

        # Convert the numpy array to a PyTorch tensor ==> internally, the data is stored as a PyTorch tensor
        if isinstance(data, ndarray):
            data = torch.from_numpy(data)

        if data is None:
            self.nodes[key] = DataNode(name)
        else:
            self.nodes[key] = DataNode(name, data)

    def add_datasets(self, path: str, **datasets: Tensor | ndarray | None):
        """Adds multiple datasets to a specified path in the DataContainer.

        Args:
            path (str): The path to the parent group.
            **datasets (Dict[str, Optional[Tensor | ndarray]]): The datasets to add, provided as a dictionary.
                The specified key will become the name of the dataset.

        Raises:
            KeyError: If the parent group does not exist.
            KeyError: If any of the datasets already exist.
        """
        for name, data in datasets.items():
            self.add_dataset(path, name, data)

    def get_dataset(self, path: str) -> Tensor:
        """Get a single dataset from a specified path in the DataContainer.

        Args:
            path (str): The path to the dataset

        Returns:
            Tensor: The data stored in the dataset.

        Raises:
            KeyError: If the dataset does not exist.
            KeyError: If the node is not a dataset.
        """
        return self.nodes(path, DataNode).data

    def get_datasets(self, *paths: str) -> tuple[Tensor, ...]:
        """Get multiple datasets from specified paths in the DataContainer.

        Args:
            *paths (str): Variable number of paths to the datasets.
                Can be provided as separate arguments or unpacked from a list.

        Returns:
            Tuple[Tensor, ...]: The tensors stored in the datasets, in the same order as the input paths.

        Raises:
            KeyError: If a dataset does not exist.
            KeyError: If the node is not a dataset.
        """
        return tuple(self.get_dataset(path) for path in paths)

    def get_all_dataset_names(self) -> list[str]:
        """Get a list of all dataset names in the DataContainer.

        Returns:
            List[str]: A list of names of all datasets in the DataContainer, without their full paths.
        """
        return [node.name for node in self.nodes.values() if isinstance(node, DataNode)]

    def get_all_dataset_paths(self) -> list[str]:
        """Get a list of all dataset paths in the DataContainer.

        Returns:
            List[str]: A list of paths to all datasets in the DataContainer.
        """
        return [path for path, node in self.nodes.items() if isinstance(node, DataNode)]

    def remove_dataset(self, path: str):
        """Removes a single dataset from a specified path in the DataContainer.

        Args:
            path (str): The path to the dataset

        Raises:
            KeyError: If the dataset does not exist.
        """
        # Remove the dataset only if it is a DataNode
        if not self._is_datanode(path):
            raise KeyError(f"Dataset at path: '{path}' does not exist or is not a dataset.")
        del self.nodes[path]

    def update_dataset(self, path: str, data: Tensor | ndarray):
        """Updates a single dataset at a specified path in the DataContainer.

        Args:
            path (str): The path to the dataset.
            data (Tensor): The new data to store in the dataset.

        Raises:
            KeyError: If the dataset does not exist.
        """
        # Convert the numpy array to a PyTorch tensor ==> internally, the data is stored as a PyTorch tensor
        if isinstance(data, ndarray):
            data = torch.from_numpy(data)

        self.nodes(path, DataNode).data = data

    def update_datasets(self, *updates: tuple[str, Tensor | ndarray]):
        """Update multiple datasets in the DataContainer.

        Args:
            *updates (Tuple[str, Tensor | ndarray]): Variable number of (path, data) tuples.
                Can be provided as separate tuples or unpacked from a list.

        Raises:
            KeyError: If a dataset does not exist.
        """
        for path, data in updates:
            self.update_dataset(path, data)



from io import BytesIO

import torch

from .attribute_ops import AttributeOps
from .dataset_ops import DatasetOps
from .group_ops import GroupOps
from .node import RootNode
from .serialization_ops import DeserializationOps, SerializationOps
from .utils import is_datanode, is_groupnode
from .visualization_ops import VisualizationOps


class DataContainer(SerializationOps, DeserializationOps, VisualizationOps, GroupOps, DatasetOps, AttributeOps):
    """Manages and serializes data into HDF5 format.

    This class manages data in a hierarchical structure, similar to a HDF5 File. It provides methods to add
    groups, datasets and attributes to the data structure. The data structure can be serialized to a HDF5 file and
    deserialized from a HDF5 file using save_to_hdf5 and load_from_hdf5 methods respectively. It also provides methods
    for visualization of the data structure.
    """

    def __init__(self, hdf5_data: BytesIO | None = None):
        """Initializes a DataContainer instance.

        By default, initializes an empty DataContainer.
        If a serialized HDF5 file is provided, the DataContainer is initialized with the data from the BytesIO object.

        Args:
            hdf5_data (BytesIO | None): The HDF5 file to deserialize. Defaults to None.
        """
        super().__init__()

        # Add the root node to the data structure
        self.nodes["/"] = RootNode()

        # If provided, initialize from a serialized HDF5 file.
        if hdf5_data:
            self.deserialize(hdf5_data)

    # Overwrite the __str__ method to provide a string representation of the DataContainer
    def __str__(self):
        returnstring = ""
        for path, node in self.nodes.items():
            returnstring = returnstring + f"{path}: ({node.name}: {node.type})" + "\n"

        return returnstring

    # Overwrite the __eq__ method to provide a comparison between two DataContainer instances
    def __eq__(self, other: object) -> bool:  # pylint: disable=too-many-return-statements
        """Compare two DataContainers for equality.

        Containers are equal if they have:
        1. Same node structure
        2. Equal data in all datasets
        3. Identical attributes for every group and dataset

        **Note**: This implements strict equality. Even with the same initial data, containers
        that have undergone stochastic transforms (e.g. GaussianNoise) will not be equal
        since their data differs.

        Args:
            other (object): The other object to compare with.

        Returns:
            bool: True if the two DataContainers are equal, False otherwise.
        """
        # Check if the other object is an instance of DataContainer
        if not isinstance(other, DataContainer):
            return False

        # Check if node structure is equal
        if set(self.nodes.keys()) != set(other.nodes.keys()):
            return False

        # Compare each node
        for path, node in self.nodes.items():
            # Retrieve other node
            other_node = other.nodes[path]

            # Compare node types and names
            if node.type != other_node.type or node.name != other_node.name:
                return False

            # Checks for GroupNodes
            if is_groupnode(node) and is_groupnode(other_node):
                # Compare attributes
                if dict(node.attributes) != dict(other_node.attributes):
                    return False

            # Checks for DataNodes
            if is_datanode(node) and is_datanode(other_node):
                # Compare attributes
                if dict(node.attributes) != dict(other_node.attributes):
                    return False

                # Compare actual data
                if not torch.equal(node.data, other_node.data):
                    return False
        return True



from abc import ABC, abstractmethod
from collections.abc import ItemsView
from enum import Enum
from sys import getsizeof

import objsize
import torch
from torch import Tensor

from ..units import Unit

AttributeTypes = str | int | float | list | tuple | dict | Unit


class NodeType(Enum):
    ROOT = "root"
    DATASET = "dataset"
    GROUP = "group"


class BaseNode(ABC):
    @abstractmethod
    def __init__(self, name: str, node_type: NodeType) -> None:
        self.__name: str = name
        self.__node_type: NodeType = node_type

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str) -> None:
        self.__name = name

    @property
    def type(self) -> NodeType:
        return self.__node_type

    def memory_bytes(self) -> int:
        """Returns the memory size of the node in bytes."""
        return getsizeof(self) + getsizeof(self.__name) + getsizeof(self.__node_type)


class RootNode(BaseNode):
    def __init__(self):
        super().__init__("root", NodeType.ROOT)


class AttributeNode(BaseNode, ABC):
    @abstractmethod
    def __init__(self, name: str, node_type: NodeType) -> None:
        super().__init__(name, node_type)
        self.__attributes: dict[str, AttributeTypes] = {}

    @property
    def attributes(self) -> ItemsView[str, AttributeTypes]:
        return self.__attributes.items()

    def add_attribute(self, key: str, value: AttributeTypes):
        if key in self.__attributes:
            raise KeyError(
                f"Attribute with key '{key}' in node '{self.name}' already exists. "
                "Use 'update_attribute' to update the value."
            )
        self.__attributes[key] = value

    def add_attributes(self, **attributes: AttributeTypes):
        for key, value in attributes.items():
            self.add_attribute(key, value)

    def get_attribute(self, key: str) -> AttributeTypes:
        if key not in self.__attributes:
            raise KeyError(f"Attribute with key '{key}' in node '{self.name}' does not exist.")
        return self.__attributes[key]

    def remove_attribute(self, key: str) -> None:
        if key not in self.__attributes:
            raise KeyError(f"Attribute with key '{key}' in node '{self.name}' does not exist.")
        del self.__attributes[key]

    def update_attribute(self, key: str, value: AttributeTypes) -> None:
        if key not in self.__attributes:
            raise KeyError(
                f"Attribute with key '{key}' in node '{self.name}' does not exist. "
                f"Use 'add_attribute' to add a new attribute."
            )

        if type(value) is not type(self.__attributes[key]):
            raise TypeError(
                f"Attribute with key '{key}' in node '{self.name}' is of type '{type(self.__attributes[key])}'. "
                f"Cannot update attribute with value of type '{type(value)}'."
            )
        self.__attributes[key] = value

    def update_attributes(self, **attributes: AttributeTypes):
        for update_key, update_value in attributes.items():
            self.update_attribute(update_key, update_value)

    def clear_attributes(self) -> None:
        self.__attributes.clear()

    def memory_bytes(self) -> int:
        """Returns the memory size of the node in bytes."""
        return super().memory_bytes() + objsize.get_deep_size(self.__attributes)


class GroupNode(AttributeNode):
    def __init__(self, name: str):
        super().__init__(name, NodeType.GROUP)


class DataNode(AttributeNode):
    def __init__(self, name: str, data: Tensor | None = None):
        super().__init__(name, NodeType.DATASET)
        self.__data: Tensor = data if data is not None else torch.empty(0)

    @property
    def data(self) -> Tensor:
        return self.__data

    @data.setter
    def data(self, value: Tensor) -> None:
        self.__data = value

    @data.deleter
    def data(self) -> None:
        self.__data = torch.empty(0)

    def memory_bytes(self) -> int:
        """Returns the memory size of the node in bytes."""
        return super().memory_bytes() + getsizeof(self.__data) + self.data.element_size() * self.data.numel()



import os
from abc import ABC, abstractmethod
from multiprocessing.pool import ThreadPool

from tqdm.auto import tqdm

from ..config import settings
from ..data import DataContainer
from ..data.datacontainer.serialization_ops import CompressionType
from ..io.backends import BaseBackend
from ..readers.base_reader import BaseReader


class BaseWriter(ABC):
    def __init__(self):
        """Constructor for the BaseWriter class. Should be called by all subclasses."""
        # Internal state
        self.__backend: BaseBackend | None = None

    @property
    def backend(self) -> BaseBackend:
        """The backend that the writer uses to write the data."""
        if not self.__backend:
            self.__backend = self._create_backend()
        return self.__backend

    @abstractmethod
    def write(
        self,
        container: DataContainer,
        file_name: str,
        compression: CompressionType = "lzf",
        compression_opts: int | None = 4,
    ):
        """Actual implementation of the writing a single DataContainer to the destination folder.

        Args:
            container (DataContainer): The DataContainer which should be written to the destination folder.
            file_name (str): The name of the DataContainer.
            compression (CompressionType): The compression method to use for the HDF5 file.
                Default is "lzf" which is a fast compression method. For smaller files, "gzip" can be used at the cost
                of speed. Use "none" to disable compression for faster read/write operations, resulting in larger files.
            compression_opts (int): The compression level for gzip compression. Ignored if compression is not "gzip".
                Default is 4, which is a good balance between speed and compression ratio.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _create_backend(self) -> BaseBackend:
        """Create a new backend instance.

        This method must be implemented by subclasses to create or
        recreate their backend when needed or after unpickling.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def process_parallel(
        self,
        reader: BaseReader,
        keep_file_names: bool = False,
        file_name_pattern: str = "{index}",
        compression: CompressionType = "lzf",
        compression_opts: int | None = 4,
        num_workers: int | None = None,
    ):
        """Process multiple DataContainers from a reader in parallel.

        Args:
            reader: Reader containing DataContainers to write
            keep_file_names: Whether to keep the original file names from the reader. If True, the argument
                `file_name_pattern` will be ignored.
            file_name_pattern: Pattern for naming files. Use {index} for zero-padded index. If {index} is not present,
                it will be appended to the pattern with an underscore.
                Example: "data_{index}_name" produces "data_00000_name.hdf5", "data_00001_name.hdf5", etc.
            compression: Compression method for HDF5 files
            compression_opts: Compression level for gzip (ignored for other methods)
            num_workers: Number of workers. Defaults to global config setting.
        """
        # Determine length to format zero-padded indices
        n = len(reader)
        index_width = len(str(n))

        if "{index}" not in file_name_pattern:
            file_name_pattern += "_{index}"

        if keep_file_names:
            _ = reader.file_names  # Access file names to ensure they are loaded

        def write_single(idx: int):
            container = reader[idx]
            if keep_file_names:
                file_name = os.path.splitext(reader.file_names[idx])[0]  # Remove original extension
                if not file_name:
                    raise ValueError(f"Invalid file name at index {idx}: '{reader.file_names[idx]}'")
            else:
                # Replace {index} with zero-padded index
                file_name = file_name_pattern.replace("{index}", str(idx).zfill(index_width))
            self.write(container, file_name, compression, compression_opts)

        # Use ThreadPool for writing in parallel ==> I/O bound task
        workers = max(num_workers, 1) if num_workers is not None else settings.num_workers
        desc = f"{self.__class__.__name__} - Writing files with {workers} workers"
        if workers > 1:
            with ThreadPool(processes=workers) as pool:
                list(tqdm(pool.imap(write_single, range(n)), total=n, desc=desc, unit="files"))
        else:
            list(map(write_single, tqdm(range(n), desc=desc, unit="files")))



from ...data import DataContainer
from ..utils import IOPathWrapper
from .base_parser import BaseParser


class HDF5Parser(BaseParser):
    supported_extensions = (".h5", ".hdf5")

    @staticmethod
    def parse(data: IOPathWrapper) -> DataContainer:
        """Parses the data from the given IOPathWrapper object into a DataContainer object.

        The IOPathWrapper object must contain a serialized DataContainer object in HDF5 format.

        Args:
            data (IOPathWrapper): IOPathWrapper object containing the data to be parsed.

        Returns:
            DataContainer: The parsed data as a DataContainer object.

        Raises:
            ValueError: If the given IOPathWrapper object is empty or does not contain a valid HDF5 file.
        """
        # Create a new DataContainer from the BytesIO object and return it
        return DataContainer(data.file_obj)



import json
from collections.abc import ItemsView
from dataclasses import fields
from io import BytesIO
from typing import Any, Literal, TypeAlias

import h5py
import numpy as np

from ..units import Unit
from .attribute_ops import AttributeOps
from .base import BaseOps
from .dataset_ops import DatasetOps
from .group_ops import GroupOps
from .node import AttributeTypes, DataNode, GroupNode, RootNode
from .utils import validate_path

CompressionType: TypeAlias = Literal["lzf", "gzip", "none"]


class SerializationOps(BaseOps):
    def serialize_to_hdf5(self, compression: CompressionType = "lzf", compression_opts: int | None = 4) -> BytesIO:
        """Serializes the DataContainer instance to an HDF5 file.

        Args:
            compression (CompressionType): The compression method to use for the HDF5 file.
                Default is "lzf" which is a fast compression method. For smaller files, "gzip" can be used at the cost
                of speed. Use "none" to disable compression for faster read/write operations, resulting in larger files.
            compression_opts (int): The compression level for gzip compression. Ignored if compression is not "gzip".
                Default is 4, which is a good balance between speed and compression ratio.

        Returns:
            BytesIO: The serialized DataContainer instance as a BytesIO object.
        """
        hdf5_bytes = BytesIO()

        with h5py.File(hdf5_bytes, "w") as f:
            for path, node in self.nodes.items():
                # 1.) HDF5 file has no root node ==> can be skipped
                if isinstance(node, RootNode):
                    continue

                # 2.) If the node is a group, create a group in the HDF5 file and add attributes
                if isinstance(node, GroupNode):
                    group = f.create_group(path)
                    self._add_attributes(group, node.attributes)

                # 3.) If the node is a dataset, create a dataset in the HDF5 file and add attributes
                elif isinstance(node, DataNode):
                    # Get the data from the node and convert to NumPy array ==> h5py does not support PyTorch tensors
                    # Force=True in case the tensor is on the GPU
                    array = node.data.numpy(force=True)

                    # Create the dataset in the HDF5 file and add attributes
                    dataset = f.create_dataset(
                        path,
                        data=array,
                        compression=None if compression == "none" else compression,
                        compression_opts=compression_opts if compression == "gzip" else None,
                        fletcher32=True,
                    )
                    self._add_attributes(dataset, node.attributes)

                # 4.) If the node is neither a group nor a dataset, raise an error
                else:
                    raise TypeError(f"Node type '{type(node)}' is not supported for serialization.")

        # Reset the BytesIO object to the beginning and return it
        hdf5_bytes.seek(0)
        return hdf5_bytes

    def _add_attributes(self, h5obj: h5py.Dataset | h5py.Group, attributes: ItemsView[str, AttributeTypes]):
        """Adds attributes to an HDF5 object (group or dataset).

        Args:
            h5obj (h5py.Group or h5py.Dataset): The HDF5 object to add attributes to.
            attributes (Dict[str, str | int | float | list | tuple | dict]): The attributes to add.
        """
        for key, value in attributes:
            # Convert lists and dictionaries to JSON strings ==> HDF5 does not support these types natively
            if isinstance(value, (list, dict, Unit)):
                value = json.dumps(value.to_dict() if isinstance(value, Unit) else value)

            # Assign attribute to HDF5 object
            h5obj.attrs[key] = value

    def save_to_hdf5(self, path: str, compression: CompressionType = "lzf", compression_opts: int | None = 4):
        """Saves the serialized DataContainer to an HDF5 file at the specified path.

        Args:
            path (str): The path where the HDF5 file should be saved.
            compression (CompressionType): The compression method to use for the HDF5 file.
                Default is "lzf" which is a fast compression method. For smaller files, "gzip" can be used at the cost
                of speed. Use "none" to disable compression for faster read/write operations, resulting in larger files.
            compression_opts (int): The compression level for gzip compression. Ignored if compression is not "gzip".
                Default is 4, which is a good balance between speed and compression ratio.
        """
        with open(path, "wb") as file:
            file.write(self.serialize_to_hdf5(compression, compression_opts).getvalue())


class DeserializationOps(GroupOps, DatasetOps, AttributeOps):
    def deserialize(self, hdf5_data: BytesIO):
        """Deserializes an HDF5 file into the DataContainer instance.

        Args:
            hdf5_data (BytesIO): The HDF5 file to deserialize.
        """
        # Check if the IOPathWrapper object is empty
        if hdf5_data.getbuffer().nbytes == 0:
            raise ValueError("The given IOPathWrapper object is empty.")

        # Check if the IOPathWrapper object is a HDF5 file
        try:
            with h5py.File(hdf5_data, "r") as f:
                pass
        except OSError as o:
            raise ValueError("The given IOPathWrapper object does not contain a valid HDF5 file.") from o

        # Check if a root node exists in the DataContainer ==> if not, add it
        if not self._is_rootnode("/"):
            self.nodes["/"] = RootNode()

        # Open the HDF5 file in read mode
        with h5py.File(hdf5_data, "r") as f:
            # Iterate over all items in the HDF5 file
            for item_name, item in f.items():
                if isinstance(item, h5py.Group):
                    self._process_group("/", item_name, item)

                elif isinstance(item, h5py.Dataset):
                    self._process_dataset("/", item_name, item)

                else:
                    raise TypeError(f"Item type '{type(item)}' is not supported for deserialization.")

    def _process_group(self, path: str, name: str, group: h5py.Group):
        """Processes and adds one group and its subgroups and datasets to the DataContainer.

        Args:
            path (str): The path to the parent where the group should be added.
            name (str): The name of the group.
            group (h5py.Group): The group to process.
        """
        # Add the group to the DataContainer
        self.add_group(path, name)

        # If the group has attributes, add them to the group in the DataContainer
        if group.attrs:
            self._process_attributes(f"{path}/{name}", dict(group.attrs))

        # Iterate over all items in the group
        for item_name, item in group.items():
            if isinstance(item, h5py.Group):
                self._process_group(f"{path}/{name}", item_name, item)

            elif isinstance(item, h5py.Dataset):
                self._process_dataset(f"{path}/{name}", item_name, item)

            else:
                raise TypeError(f"Item type '{type(item)}' is not supported for deserialization.")

    def _process_dataset(self, path: str, name: str, dataset: h5py.Dataset):
        """Processes and adds a dataset to the DataContainer.

        Args:
            path (str): The path to the parent group where the dataset should be added.
            name (str): The name of the dataset.
            dataset (h5py.Dataset): The dataset to process.
        """
        # Add data to dataset
        self.add_dataset(path, name, dataset[()])

        # Add dataset attributes if they exist
        if dataset.attrs:
            self._process_attributes(f"{path}/{name}", dict(dataset.attrs))

    def _process_attributes(self, path: str, attributes: dict[str, Any]):
        """Processes and adds attributes to a node in the DataContainer.

        Args:
            path (str): The path to the node.
            attributes (Dict[str, Any]): The attributes to add to the node.
        """
        # validate the path using the utility function
        path = validate_path(path)

        # Iterate over all attributes
        for key, value in attributes.items():
            # Convert JSON strings back to lists and dictionaries
            # Try to convert the string to a list or dictionary using the JSON format
            try:
                if isinstance(value, str):
                    value_decoded = json.loads(value)

                    # Use the JSON-decoded value only if it successfully decodes to a list or dictionary
                    if isinstance(value_decoded, (list, dict)):
                        if isinstance(value_decoded, dict) and value_decoded.keys() == {f.name for f in fields(Unit)}:
                            # Convert dict to Unit instance
                            value = Unit(**value_decoded)
                        else:
                            value = value_decoded

            # If it fails, the value is not a JSON object or a string ==> keep it as it is
            except (json.JSONDecodeError, TypeError):
                pass

            # Convert NumPy scalars to Python native types
            if isinstance(value, np.number):
                # This handles all numpy scalar types (float64, int64, etc.)
                value = value.item()

            # Add the attribute to the node in the DataContainer
            self.add_attribute(path, key, value)

    def load_from_hdf5(self, path: str):
        """Loads an HDF5 file from the specified path and deserializes it into the DataContainer.

        Args:
            path (str): The path of the HDF5 file to load.
        """
        with open(path, "rb") as file:
            hdf5_bytes = file.read()
        self.deserialize(BytesIO(hdf5_bytes))



import tarfile
import xml.etree.ElementTree as ET
from collections.abc import Sequence
from enum import IntEnum
from xml.etree.ElementTree import Element

import torch

from ...data import DataContainer, ThermoContainer
from ...data.units import arbitrary, hertz
from ...io.utils import IOPathWrapper
from .base_parser import BaseParser

TAR_HEADER_SIZE = 512  # Size of the TAR header in bytes ==> needs to be skipped when reading the data


class DataType(IntEnum):
    SHEAROGRAPHY_IMAGE = 0
    INTENSITY_IMAGE = 2
    TEMPERATURE_IMAGE = 5
    COMPLEX_IMAGE = 13


class EdevisParser(BaseParser):
    supported_extensions = (".di", ".OTvis", ".ITvisPulse")

    @staticmethod
    def parse(data: IOPathWrapper) -> DataContainer:  # pylint: disable=too-many-locals, too-many-statements, too-many-branches
        """Parses the data from the given IOPathWrapper object into a DataContainer object.

        The IOPathWrapper object must contain a .di or .OTvis file with Measurement data from Edevis.

        Parameters:
            data (IOPathWrapper): IOPathWrapper object containing the data to be parsed.

        Returns:
            DataContainer: The parsed data as a DataContainer object.

        Raises:
            ValueError: If the file in the given IOPathWrapper object is corrupted
        """
        # Extract the file object from the IOPathWrapper
        data_bytes = data.file_obj

        # Check if the BytesIO object is empty
        if data_bytes.getbuffer().nbytes == 0:
            raise ValueError("The given BytesIO object is empty.")

        # Create an empty ThermoContainer - with predefined structure
        container = ThermoContainer()

        # Add source as an attribute
        container.add_attributes(path="/MetaData", Source="Measurement: Edevis")

        # Parse data
        try:
            # Open file in tar mode and extract content.xml
            with tarfile.open(fileobj=data_bytes, mode="r:", ignore_zeros=True) as tar_file:
                # Check if content.xml exists
                try:
                    content_extracted = tar_file.extractfile("content.xml")
                except KeyError as e:
                    raise ValueError("File seems corrupted! content.xml not found.") from e

                if content_extracted is None:
                    raise ValueError("File seems corrupted! content.xml not found.")

                # Parse the extracted content.xml
                content = ET.parse(content_extracted)

                # Extract metadata from FileInfo
                file_info = content.find("FileInfo")
                if file_info is None:
                    raise ValueError("File seems corrupted! No FileInfo node found.")

                excitation_fields = [
                    "ExcitationSignalType",
                    "ExcitationFrequency",
                    "ExcitationDelay",
                    "ExcitationPulseLength",
                    "ExcitationAmplitude",
                    "ExcitationPeriods",
                    "ExcitationPhaseShift",
                    "ExcitationOnTime",
                    "ExcitationAsymmetry",
                    "ExcitationArbitrarySource",
                ]

                # Extract target fields from FileInfo
                target_fields = [
                    "UniqueIdentifier",
                    "CreationDate",
                    "CreatingVersion",
                    "ModuleName",
                    "ModuleType",
                    "SensorType",
                    "MeasurementMode",
                    "IntegrationTime",
                    "CameraSynchronization",
                    "ExcitationDeviceSelection",
                    *excitation_fields,
                ]
                metadata = extract_metadata_from_xml(file_info, target_fields)

                # Process the Sequence data
                sequences = {int(seq.attrib.get("id", -1)): seq for seq in content.findall("Sequence")}

                # TODO: For now only the first sequence is processed ==> Extend to support multiple sequences in future
                sequences = {seq_id: seq for seq_id, seq in sequences.items() if seq_id == 0}

                if len(sequences) == 0:
                    raise ValueError("File seems corrupted! No Sequence node found.")

                for seq_id, sequence in sequences.items():
                    # Add the sequence ID to the metadata
                    metadata["SequenceID"] = seq_id

                    # Process sequence info
                    sequence_info = sequence.find("SequenceInfo")
                    if sequence_info is None:
                        raise ValueError(f"Sequence {seq_id} seems corrupted! No SequenceInfo node found.")

                    target_fields = [
                        "CameraManufacturer",
                        "CameraName",
                        "Lens",
                        "FrameCount",
                        "FrameRate",
                        "Window",
                        "MaxFrameRate",
                        "MaxWindow",
                        "DataType",
                        "BitDepth",
                        "IntegrationTime",
                        *excitation_fields,
                    ]
                    metadata.update(extract_metadata_from_xml(sequence_info, target_fields))

                    # Determine width and height from the sequence info
                    if "Window" not in metadata:
                        raise ValueError(f"Sequence {seq_id} seems corrupted! No Window node found in SequenceInfo.")
                    # Window is in format "X_start,Y_Start, Width,Height"
                    width = int(metadata["Window"].split(",")[2])
                    height = int(metadata["Window"].split(",")[3])

                    # Get frame info
                    frame_info = sequence.find("FrameInfo")
                    if frame_info is None:
                        raise ValueError(f"Sequence {seq_id} seems corrupted! No FrameInfo node found.")

                    # Find all frames in FrameInfo node
                    frames = frame_info.findall("Frame")
                    if not frames:
                        raise ValueError(f"Sequence {seq_id} seems corrupted! No Frame nodes found in FrameInfo.")

                    # Get DataType and BitDepth to determine how to process the data
                    data_type = int(metadata.get("DataType", -1))
                    bit_depth = int(metadata.get("BitDepth", -1))

                    # Extract the LUT if available
                    # TODO: Test lut extraction
                    tar_offset_lut = None
                    if "TarFileHeaderCalibrationOffset" in sequence_info.attrib:
                        tar_offset_lut = int(sequence_info.attrib["TarFileHeaderCalibrationOffset"])

                        # Extract and store the LUT if available
                        if tar_offset_lut and bit_depth == 16:
                            # Reset file position to start
                            data_bytes.seek(0)

                            # Skip to LUT position
                            data_bytes.seek(tar_offset_lut + TAR_HEADER_SIZE)

                            # Extract LUT data
                            lut_size = 2**16  # If bit_depth is 16 ==> UINT16 ==> 65536 entries
                            buffer = bytearray(data_bytes.read(lut_size * 4))  # 4 bytes per LUT entry assuming float32
                            lut_data = torch.frombuffer(buffer, dtype=torch.float32)

                            # Convert LUT data to Kelvin because Thermocontainer stores LUT in Kelvin
                            container.update_dataset("/MetaData/LookUpTable", lut_data + 273.15)

                    # Define supported bit depths for each data type
                    supported_bit_depths = {
                        13: [32, 64, 128],  # Complex images
                        "default": [16, 32, 64],  # Other data types
                    }

                    # Validate bit depth
                    valid_depths = supported_bit_depths.get(data_type, supported_bit_depths["default"])
                    if bit_depth not in valid_depths:
                        raise ValueError(
                            f"Unsupported BitDepth: {bit_depth} for DataType: {data_type}. "
                            f"Supported values are {valid_depths}."
                        )

                    # Map bit depth to corresponding torch data type
                    if data_type == DataType.COMPLEX_IMAGE:  # Complex images need complex dtypes
                        frame_dtype = {32: torch.complex32, 64: torch.complex64, 128: torch.complex128}
                    else:
                        frame_dtype = {16: torch.uint16, 32: torch.float32, 64: torch.float64}

                    # Each frame is stored in a separate file. Because of the block size of the tar file,
                    # we need to skip the header size of the tar file (512 bytes) and the size of the frame header
                    # The frame header size is not explicitly given, so we need to calculate it based on the frame size
                    # Get the size of the first frame to be able to dynamically calculate the frame header size
                    first_idx = frames[0].findtext("FrameIndex", default=None)
                    try:
                        file_size = tar_file.getmember(f"sequence{seq_id}/f{first_idx}.bin").size
                    except KeyError as e:
                        msg = f"Frames in Sequence {seq_id} seem corrupted! Frame file f{first_idx}.bin not found."
                        raise ValueError(msg) from e

                    # Calculate bytes per pixel and frame size
                    bytes_per_pixel = bit_depth // 8
                    frame_size_bytes = width * height * bytes_per_pixel

                    # Dynamically calculate the frame header size
                    header_size = file_size - frame_size_bytes

                    # Pre-allocate variables and arrays for better performance
                    domain_unit = None
                    frame_unit = None
                    num_frames = len(frames)
                    domain_values = torch.zeros(num_frames, dtype=torch.float32)
                    frame_data = torch.zeros((height, width, num_frames), dtype=frame_dtype[bit_depth])

                    for i, frame in enumerate(frames):
                        # Extract frame attributes
                        offset = int(frame.attrib.get("TarFileHeaderDataOffset", -1))

                        # Handle different data types
                        match data_type:
                            case DataType.SHEAROGRAPHY_IMAGE:
                                raise NotImplementedError("Shearography Image data type is not implemented yet.")
                            case DataType.INTENSITY_IMAGE:
                                raise NotImplementedError("Intensity Image data type is not implemented yet.")
                            case DataType.TEMPERATURE_IMAGE:
                                raise NotImplementedError("Temperature Image data type is not implemented yet.")
                            case DataType.COMPLEX_IMAGE:
                                # Read domain values from the data
                                domain_str = frame.findtext("FourierFrequency", default=None)
                                domain_unit = hertz
                                if domain_str:
                                    domain_values[i] = float(domain_str.strip().split("Hz")[0])

                                # Read frame data
                                data_bytes.seek(offset + TAR_HEADER_SIZE + header_size)
                                buffer = bytearray(data_bytes.read(frame_size_bytes))
                                frame_data[:, :, i] = torch.frombuffer(buffer, dtype=frame_dtype[bit_depth]).reshape(
                                    height, width
                                )
                                frame_unit = arbitrary  # Complex images do not have a specific unit

                            case _:
                                types = ", ".join(f"{t.value} ({t.name})" for t in DataType)
                                raise ValueError(f"Unsupported DataType: {data_type}. Supported values are {types}.")

                    # Update container
                    if domain_unit:
                        container.update_dataset("/MetaData/DomainValues", domain_values)
                        container.update_unit("/MetaData/DomainValues", domain_unit)
                    if frame_unit:
                        container.update_dataset("/Data/Tdata", frame_data)
                        container.update_unit("/Data/Tdata", frame_unit)

                # # Create a simple excitation signal based on pulse length and domain values
                # # TODO: This is a simplified approach - actual signal might need more processing
                # excitation_signal = np.zeros_like(domain_values)
                # if data_type == 2:  # Only for time domain
                #     # Set signal to 1 for the duration of the pulse
                #     excitation_signal[domain_values <= pulse_length] = 1

                # container.update_dataset("/MetaData/ExcitationSignal", excitation_signal)

                # Add metadata to the container
                if metadata:
                    container.add_attributes("/MetaData", **metadata)

                # Return the container
                return container

        except Exception as e:
            raise ValueError(f"Error parsing Edevis file: {e}") from e


def extract_metadata_from_xml(xml_root: Element, target_fields: Sequence[str] | None) -> dict:
    """Extracts the target fields from the XML root element.

    Will iterate through all the children of the XML root element and extract the text
    for the specified target fields. Fields that are not found will be skipped without error.

    Args:
        xml_root (Element): The root element of the XML document.
        target_fields (Sequence[str] | None): A sequence of target field names to extract.
            If None, all fields will be extracted.

    Returns:
        dict[str, str]: A dictionary containing the extracted field names and their corresponding text values.
    """
    target = set(target_fields) if target_fields else set()  # Convert to set for faster lookup
    metadata = {}

    # If target is not that long, directly calling find is more efficient
    if 0 < len(target) < 10:
        for field_name in target:
            element = xml_root.find(field_name)
            if element is not None and element.text:
                metadata[field_name] = element.text.strip()

    # For longer target lists, or when target is empty, iterate through all children
    else:
        for child_element in xml_root:
            if (child_element.tag in target or len(target) == 0) and child_element.text:
                metadata[child_element.tag] = child_element.text.strip()

            # Stop iteration if all target fields are found
            if len(metadata) == len(target):
                break

    return metadata



import math
from collections.abc import Sequence

import torch

from ..data import DataContainer
from ..data.units import dimensionless, hertz
from .base import ThermoTransform


class PulsePhaseThermography(ThermoTransform):
    """Transform thermal data from time domain to frequency domain using FFT.

    Applies FFT to temperature data and stores complex-valued frequency components
    at specified frequencies. The complex representation preserves both amplitude
    and phase information, which can be extracted as needed for analysis.

    The time-domain data at **/Data/Tdata** is replaced with complex frequency components,
    and **/MetaData/DomainValues** is updated from time (seconds) to frequency (hertz).
    Any existing **/MetaData/ExcitationSignal** is removed as it has no meaning in the
    frequency domain.
    """

    def __init__(self, freq_indices: Sequence[int] | None = None):
        """Initialize PPT transform.

        Args:
            freq_indices: Specific frequency bin indices to keep after FFT. If None, all frequencies are retained.
                Index 0 = DC component, 1 = first harmonic, etc.
        """
        super().__init__()
        self.freq_indices = list(freq_indices) if freq_indices else None

    def forward(self, container: DataContainer) -> DataContainer:
        # Extract data
        tdata, domain_values = container.get_datasets("/Data/Tdata", "/MetaData/DomainValues")

        # Verify time domain
        if container.get_unit("/MetaData/DomainValues").quantity != "time":
            raise ValueError("PulsePhaseThermography requires time-domain data")

        if domain_values.numel() < 2:
            raise ValueError("PulsePhaseThermography requires at least two time points to compute frequency spacing")

        # Compute FFT along time axis (dim=-1)
        # Use rfft to only get positive frequencies below the Nyquist limit
        fft_result = torch.fft.rfft(tdata, dim=-1)  # pylint: disable=not-callable

        # Calculate frequency bins
        dt = domain_values[1] - domain_values[0]
        # Validate uniform sampling (required for FFT frequency calculation)
        diffs = torch.diff(domain_values)
        if not torch.allclose(diffs, dt, rtol=1e-5):
            raise ValueError("Domain values must be uniformly spaced for FFT")
        freqs = torch.fft.rfftfreq(len(domain_values), d=dt.item())  # pylint: disable=not-callable

        # Select specific frequencies if requested
        if self.freq_indices is not None:
            # Validate indices
            max_idx = fft_result.shape[-1]
            if any(idx < 0 or idx >= max_idx for idx in self.freq_indices):
                raise ValueError(f"freq_indices must be in range [0, {max_idx - 1}], got {self.freq_indices}")

            fft_result = fft_result[..., self.freq_indices]
            freqs = freqs[self.freq_indices]

        # Update container with complex FFT results
        container.update_dataset("/Data/Tdata", fft_result)  # Unit stays the same
        container.update_dataset("/MetaData/DomainValues", freqs)
        container.update_unit("/MetaData/DomainValues", hertz)  # Unit is now frequency (hertz)
        # TODO: Refactor once public API of DataContainer is improved
        if container._is_datanode("/MetaData/ExcitationSignal"):  # pylint: disable=protected-access
            container.remove_dataset(
                "/MetaData/ExcitationSignal"
            )  # ExcitationSignal is meaningless in frequency domain
        return container


class ExtractAmplitude(ThermoTransform):
    """Extract amplitude images from complex thermal data."""

    def forward(self, container: DataContainer) -> DataContainer:
        tdata = container.get_dataset("/Data/Tdata")

        if not torch.is_complex(tdata):
            raise ValueError(
                "Data must be complex (FFT result). Apply PulsePhaseThermography transform first to "
                "convert time-domain data to frequency domain."
            )

        # Extract amplitude
        amplitude = torch.abs(tdata)
        container.update_dataset("/Data/Tdata", amplitude)  # Unit stays the same
        return container


class ExtractPhase(ThermoTransform):
    """Extract phase images from complex thermal data."""

    def __init__(self, unwrap: bool = True) -> None:
        """Initialize phase extraction transform.

        Args:
            unwrap: If True, the phase information is unwrapped to avoid discontinuities. Defaults to True.
        """
        super().__init__()
        self.unwrap = unwrap

    def _unwrap(self, phase: torch.Tensor, discont: float = math.pi, dim: int = -1) -> torch.Tensor:
        """Unwrap phase information along the specified dimension. Matches numpy.unwrap behavior.

        Args:
            phase: Input tensor containing phase values in radians.
            discont: Discontinuity threshold, typically set to pi.
            dim: Dimension along which to unwrap the phase. Default is the last dimension.
        """
        # Normalize dim to positive index
        dim = dim if dim >= 0 else phase.ndim + dim

        # Compute differences along specified dimension
        d = torch.diff(phase, dim=dim)
        d_mod = (d + discont) % (2 * discont) - discont

        # Handle edge case where difference is exactly -discont
        mask = (d_mod == -discont) & (d > 0)
        d_mod = torch.where(mask, torch.full_like(d_mod, discont), d_mod)

        # Compute correction and accumulate
        corr = d_mod - d
        correction = torch.cumsum(corr, dim=dim)

        # Pad on the correct dimension to account for reduced size after diff
        pad_shape = list(correction.shape)
        pad_shape[dim] = 1
        zeros = torch.zeros(pad_shape, dtype=phase.dtype, device=phase.device)
        correction = torch.cat([zeros, correction], dim=dim)

        return phase + correction

    def forward(self, container: DataContainer) -> DataContainer:
        tdata = container.get_dataset("/Data/Tdata")

        if not torch.is_complex(tdata):
            raise ValueError(
                "Data must be complex (FFT result). Apply PulsePhaseThermography transform first to "
                "convert time-domain data to frequency domain."
            )

        # Extract phase and unwrap
        phase = torch.angle(tdata)
        if self.unwrap:
            phase = self._unwrap(phase, dim=-1)
        container.update_dataset("/Data/Tdata", phase)
        container.update_unit("/Data/Tdata", dimensionless)  # Phase images are now radians (dimensionless)
        return container



import itertools
import math
import warnings
from collections.abc import Callable, Sequence
from functools import partial

import torch
from torch import Generator, default_generator

from ..data import DataContainer
from ..transforms.base import _BaseTransform
from .indexed_thermo_dataset import IndexedThermoDataset
from .thermo_dataset import ThermoDataset


def random_split(
    dataset: ThermoDataset,
    lengths: Sequence,
    transforms: Sequence[_BaseTransform | None] | None = None,
    generator: Generator = default_generator,
) -> list[IndexedThermoDataset]:
    """Split a dataset into random non-overlapping subsets of given lengths with optional transforms being applied.

    If a list of fractions that sum up to 1 is given, the lengths will be computed automatically as
    floor(frac * len(dataset)) for each fraction provided.

    After computing the lengths, if there are any remainders, 1 count will be distributed in round-robin fashion to the
    lengths until there are no remainders left.

    Args:
        dataset (Dataset): Dataset to be split
        lengths (Sequence[float]): Fractions for each split that sum up to 1.0.
        transforms (Sequence[_BaseTransform | None], optional): Optional sequence of transforms for each split.
        generator (Generator, optional): Generator used for reproducible splits.
            Per default the default generator is used.

    Returns:
        List[ThermoSubset]: List of subsets with the specified lengths and transforms

    Raises:
        ValueError: If lengths don't sum up to 1.0
        ValueError: If number of transforms doesn't match lengths

    Example:
        >>> from pythermondt import transforms as T
        >>> # Create train/val/test splits with different transforms
        >>> train_transform = T.Compose([T.ApplyLUT(), T.RandomCrop(96, 96)])
        >>> val_transform = T.Compose([T.ApplyLUT(), T.CropFrames(96, 96)])
        >>> train, val, test = random_split(dataset, [0.7, 0.2, 0.1], transforms=[train_transform, val_transform, None])
    """
    # Validate transforms if provided
    if transforms is not None and len(transforms) != len(lengths):
        raise ValueError(f"Number of transforms: {len(transforms)} must match number of splits: {len(lengths)}")

    # Validate lengths
    if not all(value >= 0 for value in lengths):
        raise ValueError("All values in lengths must be greater or equal to 0")

    # If lengths are provided as fractions and not as absolute numbers
    if math.isclose(sum(lengths), 1) and sum(lengths) <= 1:
        # Compute the lengths of the subsets
        subset_lengths: list[int] = []
        for i, frac in enumerate(lengths):
            # Defensive check to ensure fractions are between 0 and 1
            if frac < 0 or frac > 1:  # pragma: no cover
                raise ValueError(f"Fraction at index {i} is not between 0 and 1")
            n_items_in_split = math.floor(len(dataset) * frac)
            subset_lengths.append(n_items_in_split)

        # Compute remainder
        remainder = len(dataset) - sum(subset_lengths)

        # add 1 to all the lengths in round-robin fashion until the remainder is 0
        for i in range(remainder):
            idx_to_add_at = i % len(subset_lengths)
            subset_lengths[idx_to_add_at] += 1
        lengths = subset_lengths

    # Print a warning if any of the splits have a length of 0
    for i, length in enumerate(lengths):
        if length == 0:
            warnings.warn(f"Length of split at index {i} is 0. This might result in an empty dataset.", stacklevel=2)

    # Raise an error if the computed lengths don't match the length of the original dataset
    if sum(lengths) != len(dataset):
        raise ValueError(
            f"The sum of the computed subset lengths: {lengths} does not match "
            f"the length of the original dataset: {len(dataset)}"
        )

    # Generate random indices
    indices = torch.randperm(sum(lengths), generator=generator).tolist()

    # Create the subsets and return
    transforms = transforms or [None] * len(lengths)
    return [
        IndexedThermoDataset(dataset, indices[offset - length : offset], transform)
        for transform, length, offset in zip(transforms, lengths, itertools.accumulate(lengths), strict=False)
    ]


def container_collate(*paths: str) -> Callable[[Sequence[DataContainer]], tuple[torch.Tensor, ...]]:
    """Factory function for creating a collate function for DataContainer objects.

    Returns a function that extracts specified dataset paths and stacks them along batch dimension.

    Args:
        *paths (str): Variable number of dataset paths to extract (e.g. '/Data/Tdata', '/GroundTruth/DefectMask')

    Returns:
        Callable[[Sequence[DataContainer]], tuple[torch.Tensor, ...]]: A collate function that takes a batch of
            DataContainer objects and returns a tuple of tensors. The number of tensors in the tuple corresponds to the
            number of dataset paths provided.

    Raises:
        KeyError: If a specified dataset path doesn't exist in any container
        RuntimeError: If tensors have incompatible shapes for stacking
        ValueError: If empty batch is provided

    Example:
        >>> from torch.utils.data import DataLoader
        >>> collate_fn = container_collate("/Data/Tdata", "/GroundTruth/DefectMask")
        >>> dataloader = DataLoader(dataset, batch_size=32, collate_fn=collate_fn)
    """
    if not paths:
        raise ValueError("At least one path must be specified")

    return partial(_container_collate_impl, paths=paths)


def _container_collate_impl(batch: Sequence[DataContainer], paths: tuple[str, ...]) -> tuple[torch.Tensor, ...]:
    """Implementation function that processes a batch of DataContainer objects for collation.

    Args:
        batch (Sequence[DataContainer]): The batch of DataContainer objects to collate.
        paths (tuple[str, ...]): The dataset paths to extract and collate from each container.

    Returns:
        tuple[torch.Tensor, ...]: A tuple of tensors, each stacked along the batch dimension for the corresponding path.

    Raises:
        KeyError: If a specified dataset path doesn't exist in any container.
        RuntimeError: If tensors have incompatible shapes for stacking.
        ValueError: If empty batch is provided.
    """
    if not batch:
        raise ValueError("Empty batch provided - cannot collate empty sequence")

    # Use get_datasets method to efficiently extract all datasets from each container
    all_tensors = []
    for container in batch:
        try:
            # Get all datasets from this container in one call
            tensors = container.get_datasets(*paths)
            all_tensors.append(tensors)
        except KeyError as exc:
            raise KeyError(f"One or more dataset paths not found in container: {exc}") from exc

    # Stack tensors along batch dimension for each path
    result = []
    for i, path in enumerate(paths):
        try:
            # Extract tensors for this path from all containers and stack them
            result.append(torch.stack([tensors[i] for tensors in all_tensors], dim=0))
        except RuntimeError as e:
            raise RuntimeError(f"Cannot stack tensors for path '{path}': {e}") from e

    return tuple(result)



from .azure_reader import AzureBlobReader
from .base_reader import BaseReader
from .local_reader import LocalReader
from .s3_reader import S3Reader

__all__ = [
    "AzureBlobReader",
    "BaseReader",
    "LocalReader",
    "S3Reader",
]



from .backends import AzureBlobBackend, BaseBackend, LocalBackend, S3Backend
from .parsers import BaseParser, EdevisParser, HDF5Parser, SimulationParser
from .utils import IOPathWrapper

__all__ = [
    "AzureBlobBackend",
    "BaseBackend",
    "BaseParser",
    "EdevisParser",
    "HDF5Parser",
    "IOPathWrapper",
    "LocalBackend",
    "S3Backend",
    "SimulationParser",
]



import inspect
from abc import ABC, abstractmethod

from torch import nn

from ..data import DataContainer


class _BaseTransform(nn.Module, ABC):
    """Abstract base class for all transforms in PyThermoNDT.

    This class is not intended to be used directly, but rather serves as a foundation for both deterministic and
    random transforms. It initializes the module and sets up necessary configurations.
    """

    @abstractmethod
    def forward(self, container: DataContainer) -> DataContainer:
        """Abstract method that must be implemented in the sub-class.

        This method should contain the actual transformation logic.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @property
    @abstractmethod
    def is_random(self) -> bool:
        """Indicates whether the transform is random or deterministic."""
        raise NotImplementedError("Subclasses must implement this method")

    def extra_repr(self) -> str:
        """Return extra representation of transform parameters.

        This follows PyTorch's pattern but makes it automatic.
        Override this method if you need custom parameter formatting.
        """
        # Get init parameters by inspecting the constructor
        try:
            sig = inspect.signature(self.__class__.__init__)
            init_params = set(sig.parameters.keys()) - {"self"}
        except (ValueError, TypeError):
            init_params = set()

        # Get attributes that match init parameters
        params = []
        for attr_name in sorted(init_params):
            if hasattr(self, attr_name):
                value = getattr(self, attr_name)
                # Only include simple types that are likely parameters
                if isinstance(value, (int, float, str, bool, list, tuple)):
                    params.append(f"{attr_name}={value}")

        return ", ".join(params)

    def __str__(self) -> str:
        """Get a string representation of the transform."""
        class_name = self.__class__.__name__
        transform_type = "Random" if self.is_random else "Deterministic"

        extra_repr = self.extra_repr()
        if extra_repr:
            return f"{class_name}({extra_repr}) [{transform_type}]"
        return f"{class_name}() [{transform_type}]"

    # Add type annotations to __call__ method, so that the type checker can infer the correct return type.
    # Otherwise, the return type will be inferred as 'Any'.
    # __call__ does not have to be overridden, this is already implemented in the nn.Module class from PyTorch.
    # See: https://github.com/microsoft/pyright/issues/3249
    def __call__(self, container: DataContainer) -> DataContainer:
        """Override to provide proper type hints."""
        return super().__call__(container)


class ThermoTransform(_BaseTransform):
    """Abstract base class that all deterministic transforms of PyThermoNDT must inherit from.

    Initializes the module and sets up necessary configurations. These transforms are expected to produce deterministic
    outputs given the same input, thus they can safely be cached for improved performance.
    """

    @property
    def is_random(self) -> bool:
        return False


class RandomThermoTransform(_BaseTransform):
    """Abstract base class that all random/stochastic transforms of PyThermoNDT must inherit from.

    Initializes the module and sets up necessary configurations. These transforms are expected to produce random outputs
    given the same input, thus they should NOT be cached for improved performance.
    """

    @property
    def is_random(self) -> bool:
        return True



"""Tests for Settings validation in config.py."""

import pytest
from pydantic import ValidationError

from pythermondt.config import Settings


def test_num_workers_zero_raises():
    """Test that num_workers=0 raises ValidationError."""
    with pytest.raises(ValidationError, match="num_workers must be at least 1"):
        Settings(num_workers=0)


@pytest.mark.parametrize("num_workers", [-5, -1])
def test_num_workers_negative_raises(num_workers):
    """Test that negative num_workers raises ValidationError."""
    with pytest.raises(ValidationError, match="num_workers must be at least 1"):
        Settings(num_workers=num_workers)


def test_num_workers_valid():
    """Test that a valid num_workers is accepted."""
    s = Settings(num_workers=2)
    assert s.num_workers == 2


def test_invalid_log_level_raises():
    """Test that an unrecognized log_level raises ValidationError."""
    with pytest.raises(ValidationError, match="log_level must be one of"):
        Settings(log_level="INVALID")


def test_invalid_log_level_typo_raises():
    """Test that a common typo like WARN raises ValidationError."""
    with pytest.raises(ValidationError, match="log_level must be one of"):
        Settings(log_level="WARN")


def test_log_level_case_insensitive():
    """Test that log_level accepts lowercase and normalizes to uppercase."""
    s = Settings(log_level="debug")
    assert s.log_level == "DEBUG"



from collections.abc import Callable, Sequence

from ..data import DataContainer
from .base import ThermoTransform, _BaseTransform


class Compose(ThermoTransform):
    """Compose a sequence of transforms together into a single transform.

    This transform sequentially applies a list of transforms to the input container.
    """

    def __init__(self, transforms: Sequence[_BaseTransform]):
        """Compose a sequence of transforms together into a single transform.

        This transform sequentially applies a list of transforms to the input container.
        """
        super().__init__()

        # Check if all the provided transforms are valid (Thermotransforms are already callable)
        if not all(isinstance(t, _BaseTransform) for t in transforms):
            raise TypeError("Not all transforms inherit from _BaseTransform.")
        self.transforms = transforms

    def __str__(self) -> str:
        """Custom repr for Compose - no type label, cleaner format."""
        if not self.transforms:
            return "Compose([])"

        # Show transforms in a clean format
        transform_strs = [str(t) for t in self.transforms]
        if len(transform_strs) == 1:
            return f"Compose([{transform_strs[0]}])"

        # Multi-line format for multiple transforms
        transforms_str = ",\n    ".join(transform_strs)
        return f"Compose([\n    {transforms_str}\n])"

    def forward(self, container: DataContainer) -> DataContainer:
        for t in self.transforms:
            container = t(container)
        return container


class CallbackTransform(_BaseTransform):
    """A transform that constructs another transform using a callback function at runtime.

    This is useful for setting up custom transforms, where the input arguments depend on runtime information.

    Per default, randomness is inferred from the callback by executing it with a dummy container. If this fails, the
    transform is assumed to be non-random. To override this behavior, set the `is_random` argument explicitly.
    """

    def __init__(self, callback: Callable[[DataContainer], _BaseTransform], is_random: bool | None = None):
        """A transform that constructs another transform using a callback function at runtime.

        This is useful for setting up custom transforms, where the input arguments depend on runtime information.

        Per default, randomness is inferred from the callback by executing it with a dummy container. If this fails, the
        transform is assumed to be non-random. To override this behavior, set the `is_random` argument explicitly.

        Args:
            callback (Callable[[DataContainer], _BaseTransform]): A function that takes a DataContainer and returns a
                transform instance. This function will be called each time the transform is applied.
            is_random (bool | None, optional): Whether the transform is random. If None, randomness is inferred from
                the callback function, with fall to deterministic if inference fails. Defaults to None.
        """
        super().__init__()
        self.callback = callback
        self._is_random = is_random

    @property
    def is_random(self) -> bool:
        if self._is_random is not None:
            return self._is_random

        # Try to infer randomness from callback
        try:
            dummy_container = DataContainer()
            transform = self.callback(dummy_container)
            return transform.is_random
        except Exception:  # pylint: disable=broad-except
            print("Warning: Could not infer randomness from callback function. Assuming deterministic.")
            return False

    def extra_repr(self) -> str:
        """Return string representation of callback."""
        callback_name = getattr(self.callback, "__name__", "lambda")
        explicit = f", is_random={self._is_random}" if self._is_random is not None else ""
        return f"callback={callback_name}{explicit}"

    def forward(self, container: DataContainer) -> DataContainer:
        # Instantiate the transform using the callback and apply it
        transform = self.callback(container)
        return transform(container)


def split_transforms_for_caching(
    transforms: _BaseTransform | Sequence[_BaseTransform],
) -> tuple[Compose, Compose]:
    """Split any composed transforms into deterministic and random transforms.

    This function takes a sequence of transforms or a Compose object and splits them into two Compose objects:
    - Deterministic transforms
    - Random transforms and any transforms that follow them

    Nested Compose transforms are flattened before splitting, so that the
    split is applied to the individual transforms rather than the Compose container itself.

    Args:
        transforms (_BaseTransform | Sequence[_BaseTransform]): Transforms to split.

    Returns:
        tuple[Compose, Compose]:
            A tuple containing two Compose objects:
            - The first contains deterministic transforms.
            - The second contains random transforms and any transforms that follow them.
    """
    # Flatten nested Compose transforms first
    if isinstance(transforms, Compose):
        transforms = transforms.transforms
    elif isinstance(transforms, _BaseTransform):
        transforms = [transforms]
    flat_transforms = _flatten_transforms(transforms)

    # Simple split on flattened list
    deterministic = []
    random_and_after = []
    found_random = False

    for transform in flat_transforms:
        if transform.is_random:
            found_random = True

        if found_random:
            random_and_after.append(transform)
        else:
            deterministic.append(transform)

    return Compose(deterministic), Compose(random_and_after)


def _flatten_transforms(transforms: Sequence[_BaseTransform]) -> Sequence[_BaseTransform]:
    """Flatten nested Compose transforms into a single list."""
    flattened: list[_BaseTransform] = []
    for transform in transforms:
        if isinstance(transform, Compose):
            # Recursively flatten nested Compose
            flattened.extend(_flatten_transforms(transform.transforms))
        else:
            flattened.append(transform)
    return flattened



import collections
import copy
import gc
import multiprocessing as mp
import sys
from abc import ABC, abstractmethod
from multiprocessing.managers import ListProxy, SyncManager
from multiprocessing.pool import ThreadPool
from typing import Literal

from torch.utils.data import Dataset
from tqdm.auto import tqdm

from ..config import settings
from ..data import DataContainer
from ..data.datacontainer.utils import format_bytes
from ..transforms.utils import Compose, _BaseTransform, _flatten_transforms, split_transforms_for_caching

CacheMode = Literal["immediate", "lazy"]


class BaseDataset(Dataset, ABC):
    """Base class for all ThermoDatasets."""

    def __init__(self, parent: "BaseDataset | None" = None, transform: _BaseTransform | None = None):
        self.__parent = parent
        self.__transform = transform

        # Internal state for cache
        self.__manager: SyncManager | None = None
        self.__cache_built = False
        self.__cache: list | ListProxy = []
        self.__det_transforms: _BaseTransform | None = None
        self.__runtime_transforms: _BaseTransform | None = None

    @abstractmethod
    def load_raw_data(self, idx: int) -> DataContainer:
        """Load raw data at index - implemented by concrete classes."""

    @abstractmethod
    def __len__(self) -> int:
        """Return length of dataset - implemented by concrete classes."""

    @property
    @abstractmethod
    def files(self) -> list[str]:
        """Get the list of files associated with this dataset."""

    def __del__(self):
        """Ensure that the cache is released when the dataset is deleted."""
        if self.cache_built:
            self.release_cache(gc_collect=False)

    def __getitem__(self, idx: int) -> DataContainer:
        """Get an item while also applying the proper transform chain.

        Args:
            idx (int): Index of the data to retrieve

        Returns:
            DataContainer: Transformed data container
        """
        # Validate index
        if idx < 0 or idx >= len(self):
            msg = f"Index {idx} out of range."
            raise IndexError(msg + (f" Must be within [0, {len(self) - 1}]" if len(self) > 0 else " Empty dataset"))

        if self.cache_built:
            if self.__cache[idx] is None:
                # Load the item if it was not cached
                self.__cache[idx] = self._load_cache_item(idx)
            data = copy.deepcopy(self.__cache[idx])
            return self.__runtime_transforms(data) if self.__runtime_transforms else data

        # Get the data
        data = self.load_raw_data(idx) if self.parent is None else self.parent[self._map_index(idx)]

        # Apply additional transform if specified
        if self.transform:
            data = self.transform(data)  # pylint: disable=not-callable

        return data

    @property
    def parent(self) -> "BaseDataset | None":
        """Get parent dataset if available."""
        return self.__parent

    @property
    def transform(self) -> _BaseTransform | None:
        """Get the transform for this dataset."""
        return self.__transform

    @property
    def cache_built(self) -> bool:
        """Check if the cache has been built."""
        return self.__cache_built

    def _map_index(self, idx: int) -> int:
        """Hook to map the index to the parent's index. Override in subclasses if needed."""
        return idx

    def _load_cache_item(self, idx: int) -> DataContainer:
        """Load a single item and apply deterministic transforms."""
        return self.__det_transforms(self.load_raw_data(idx)) if self.__det_transforms else self.load_raw_data(idx)

    def memory_bytes(self) -> int:
        """Calculate the memory usage of this dataset.

        **Note:** If the cache has not been built, this will be small because the data is not loaded yet.

        Returns:
            int: Memory usage in bytes
        """
        container_size = sum(c.memory_bytes() for c in self.__cache if isinstance(c, DataContainer))
        return container_size + sys.getsizeof(self) + sys.getsizeof(self.__cache)

    def print_memory_usage(self):
        """Print the memory usage of this dataset."""
        print(f"{self.__class__.__name__} Overview:")
        print("-" * len(f"{self.__class__.__name__} Overview:"))
        print(f"Currently there are {sum(1 for item in self.__cache if item)} items in the cache")
        print(f"Total memory usage of the cache: {format_bytes(self.memory_bytes())}")
        print("\n")

    def get_transform_chain(self) -> _BaseTransform:
        """Walk up graph to build the complete sequence transforms for this dataset and compose it in a single one."""
        transforms: collections.deque[_BaseTransform] = collections.deque()
        current: BaseDataset | None = self

        while current is not None:
            if current.transform:
                # Flatten the transforms
                flattened = _flatten_transforms(
                    current.transform.transforms if isinstance(current.transform, Compose) else [current.transform]
                )
                transforms.extendleft(reversed(flattened))
            current = current.parent

        return Compose(list(transforms))

    def build_cache(self, mode: CacheMode = "lazy", num_workers: int | None = None):
        # fmt: off
        """Build an in-memory cache of preprocessed data for faster training.

        Automatically splits the transform pipeline at the first random transform:
        - Deterministic transforms are applied once and cached
        - Random transforms + subsequent operations run at runtime

        Platform Considerations:
            **Windows**: Use lazy mode to avoid memory issues due to cache duplication in forked processes
            **Linux**: Both modes work efficiently - choose based on workflow preference

        Args:
            mode (CacheMode): Cache building strategy
                - "lazy": Create shared cache, workers fill on-demand for faster startup (default)
                - "immediate": Build all items upfront using a ThreadPool
            num_workers (int, optional): Number of workers used for cache building. This setting only applies if `mode`
                is "immediate". If num_workers is None, the global configuration of PyThermoNDT will be used.
                If less than 1, it defaults to 1 worker. Default is None.

        Example with a common preprocessing pipeline:
            >>> train_pipeline = T.Compose([
            ...     T.ApplyLUT(),                           # Cached
            ...     T.SubtractFrame(0),                     # Cached
            ...     T.RemoveFlash(method='excitation_signal'), # Cached
            ...     T.NonUniformSampling(64),               # Cached
            ...     T.RandomFlip(p_height=0.3, p_width=0.3), # Runtime (random)
            ...     T.GaussianNoise(std=25e-3),             # Runtime (random)
            ...     T.MinMaxNormalize(),                    # Runtime (after random)
            ... ])

            >>> # Development: fast startup (default)
            >>> dataset.build_cache(mode="lazy")
            >>> loader = DataLoader(dataset, num_workers=4, persistent_workers=True)

            >>> # Production: parallel cache building (Only recommended on linux)
            >>> dataset.build_cache(mode="immediate", num_workers=8)
            >>> loader = DataLoader(dataset, num_workers=8, persistent_workers=True)
        """
        # fmt: on
        # Skip if cache already built
        if self.__cache_built:
            return

        # Get the complete transform chain and split it into deterministic and runtime transforms
        self.__det_transforms, self.__runtime_transforms = split_transforms_for_caching(self.get_transform_chain())

        # Initialize the cache based on the mode
        if mode == "immediate":
            unit = "files"
            desc = f"{self.__class__.__name__} - Building cache"
            num = len(self)
            workers = max(num_workers, 1) if num_workers is not None else settings.num_workers
            worker_fn = self._load_cache_item
            if workers > 1:
                # Use ThreadPool for immediate cache building in parallel
                with ThreadPool(processes=workers) as pool:
                    self.__cache = list(tqdm(pool.imap(worker_fn, range(num)), desc=desc, unit=unit, total=num))
            else:
                self.__cache = [self._load_cache_item(i) for i in tqdm(range(num), desc=desc, unit=unit)]
        elif mode == "lazy":
            # Choose context based on platform to avoid deprecated 'fork' in multithreaded processes (Python 3.12+):
            ctx = mp.get_context("forkserver" if sys.platform == "linux" else "spawn")
            self.__manager = ctx.Manager()
            # Create a shared list for lazy loading using a list proxy
            self.__cache = self.__manager.list([None] * len(self))
        else:
            raise ValueError(f"Invalid cache mode: {mode}. Use one of: {list(CacheMode.__args__)}.")

        self.__cache_built = True

    def release_cache(self, gc_collect: bool = True):
        """Release the in-memory cache to free up memory and release any background manager processes.

        Args:
            gc_collect (bool): Whether to run garbage collection after releasing the cache. Default is True.
        """
        # For regular lists, clear the items to free memory ==> ListProxies are handled by manager shutdown
        if not isinstance(self.__cache, ListProxy):
            self.__cache = []
        self.__det_transforms = None
        self.__runtime_transforms = None
        self.__cache_built = False

        # Ensure that the manager process is terminated
        if self.__manager:
            try:
                self.__manager.shutdown()
            except Exception:  # pylint: disable=broad-except
                pass
            finally:
                self.__manager = None

        # Garbage collect to free memory if requested
        if gc_collect:
            gc.collect()



import logging
import os
import tempfile
from io import BytesIO

logger = logging.getLogger(__name__)


class IOPathWrapper:
    """Provides unified access to file content via both path and buffer interfaces."""

    def __init__(self, source: str | BytesIO | bytes):
        """Provides unified access to file content via both path and buffer interfaces.

        Initialize with either a file path or file content.

        Args:
            source: Either a file path (str), file-like object, or bytes
        """
        self.__source = source
        self.__file_obj: BytesIO | None = None
        self.__temp_path: str | None = None

    @property
    def file_obj(self) -> BytesIO:
        """Get file-like object, loading from path if needed."""
        if self.__file_obj is None:
            if isinstance(self.__source, str):
                # Path provided - load file when first needed
                with open(self.__source, "rb") as f:
                    self.__file_obj = BytesIO(f.read())
            elif isinstance(self.__source, bytes):
                self.__file_obj = BytesIO(self.__source)
            elif isinstance(self.__source, BytesIO):
                # File-like object provided (from boto3 etc.)
                self.__file_obj = self.__source
            else:
                raise ValueError("Unsupported source type. Must be str, bytes, or BytesIO.")

        # Reset position and return
        self.__file_obj.seek(0)
        return self.__file_obj

    @property
    def file_path(self) -> str:
        """Get file path, using original path or creating temp file."""
        if isinstance(self.__source, str) and os.path.exists(self.__source):
            # Source is already a valid path - use directly
            return self.__source

        # Create temporary file if needed
        if not self.__temp_path or not os.path.exists(self.__temp_path):
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                self.file_obj.seek(0)
                temp.write(self.file_obj.getbuffer())
                self.__temp_path = temp.name

        return self.__temp_path

    def close(self):
        """Close resources and remove temporary file."""
        # Remove temp file
        if self.__temp_path and os.path.exists(self.__temp_path):
            try:
                os.remove(self.__temp_path)
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(
                    "Failed to remove temporary file %s: %s",
                    self.__temp_path,
                    e,
                    exc_info=True,  # Include traceback for debugging
                )
            self.__temp_path = None

    def __del__(self):
        """Ensure cleanup on garbage collection."""
        self.close()



import os

from ..data import DataContainer
from ..data.datacontainer.serialization_ops import CompressionType
from ..io import IOPathWrapper, LocalBackend
from .base_writer import BaseWriter


class LocalWriter(BaseWriter):
    def __init__(self, destination_folder: str):
        """Instantiates a new instance of the LocalWriter class.

        Args:
            destination_folder (str): The destination folder where the DataContainers should be written to.
        """
        super().__init__()

        # Verify folder
        self.__exists = os.path.exists(destination_folder)
        self.__destination_folder = destination_folder

    def _create_backend(self) -> LocalBackend:
        return LocalBackend(pattern=self.__destination_folder)

    def write(
        self,
        container: DataContainer,
        file_name: str,
        compression: CompressionType = "lzf",
        compression_opts: int | None = 4,
    ):
        # Verify folder
        if not self.__exists or not os.path.exists(self.__destination_folder):
            os.makedirs(self.__destination_folder, exist_ok=True)

        # Append file extension if not present
        if not file_name.endswith(".hdf5"):
            file_name += ".hdf5"

        # Create the path to the file
        path = os.path.join(self.__destination_folder, file_name)

        # Write the DataContainer to the file
        self.backend.write_file(IOPathWrapper(container.serialize_to_hdf5(compression, compression_opts)), path)



from .azure_writer import AzureBlobWriter
from .base_writer import BaseWriter
from .local_writer import LocalWriter
from .s3_writer import S3Writer

__all__ = ["AzureBlobWriter", "BaseWriter", "LocalWriter", "S3Writer"]



import logging
import os

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


def configure_logging(level: str | None = None):
    """Configure logging for pythermondt.

    Simple convenience function for quick debugging. Advanced users should
    configure Python's logging directly.

    Args:
        level: Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL).
               If None, uses settings.log_level.

    Example:
        >>> import pythermondt
        >>> pythermondt.configure_logging("DEBUG")  # See all logs
        >>> pythermondt.configure_logging()  # Use settings.log_level
    """
    # Add a basic configuration but keep loglevel at default (WARNING)
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        force=True,  # Override any existing config
    )

    # Capture warnings from the warnings module as well
    logging.captureWarnings(True)

    # Configure log level specifically for the pythermondt logger ==> avoids spamming from other libraries
    level = level or settings.log_level
    logger = logging.getLogger("pythermondt")
    logger.setLevel(getattr(logging, level.upper()))


class Settings(BaseSettings):
    """Global settings for PyThermoNDT."""

    # Configuration parameters
    download_dir: str = Field(default="./", description="Base directory where PyThermoNDT will download files to.")
    num_workers: int = Field(
        default=os.cpu_count() or 1, description="Default number of workers used for parallel operations."
    )
    log_level: str = Field(
        default="WARNING", description="Default log level for pythermondt (DEBUG, INFO, WARNING, ERROR, CRITICAL)"
    )

    model_config = SettingsConfigDict(
        env_prefix="PYTHERMONDT_",
        case_sensitive=False,
        env_file=".env",
        env_file_encoding="utf-8",
        extra="allow",
    )

    @field_validator("download_dir", mode="after")
    @classmethod
    def ensure_exists(cls, v: str) -> str:
        # Use os.path for validation/normalization
        expanded = os.path.expanduser(v)
        absolute = os.path.abspath(expanded)

        # Create if missing
        os.makedirs(absolute, exist_ok=True)

        return absolute

    @field_validator("num_workers", mode="before")
    @classmethod
    def validate_num_workers(cls, v: int) -> int:
        if v < 1:
            raise ValueError("num_workers must be at least 1")
        return v

    @field_validator("log_level", mode="after")
    @classmethod
    def validate_log_level(cls, v: str) -> str:
        valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        v_upper = v.upper()
        if v_upper not in valid_levels:
            raise ValueError(f"log_level must be one of {valid_levels}, got {v}")
        return v_upper


# Global settings instance
settings = Settings()



import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib import ticker
from matplotlib.lines import Line2D
from matplotlib.offsetbox import AnnotationBbox, TextArea
from matplotlib.widgets import Button, CheckButtons, Slider

from ..units import generate_label
from .attribute_ops import AttributeOps
from .dataset_ops import DatasetOps
from .group_ops import GroupOps


class VisualizationOps(GroupOps, DatasetOps, AttributeOps):
    _interactive_analyzer: "VisualizationOps.InteractiveAnalyzer | None" = None

    # TODO: Refactor visualization logic to reduce the tight coupling between data handling and visualization.
    class InteractiveAnalyzer:  # pylint: disable=too-many-instance-attributes
        def __init__(self, parent: "VisualizationOps"):
            """Initialize the interactive analyzer for thermographic data visualization.

            Args:
                parent (VisualizationOps): The parent container for the interactive analysis.
            """
            # 1.) Retrieve data from the container
            self.container = parent
            # Transpose to (frame, y, x) for faster access - this avoids the need to squeeze the data
            self.tdata = parent.get_dataset("/Data/Tdata").numpy(force=True).transpose(2, 0, 1)
            self.domain_values = parent.get_dataset("/MetaData/DomainValues").numpy(force=True)
            self.data_unit = parent.get_unit("/Data/Tdata")
            self.domain_unit = parent.get_unit("/MetaData/DomainValues")

            # 2.) Setup the figure, axes and colorbar
            # Create the main figure with two subplots
            self.fig = plt.figure(figsize=(15, 6))
            self.frame_ax = plt.subplot2grid((1, 2), (0, 0))
            self.profile_ax = plt.subplot2grid((1, 2), (0, 1))

            # Initialize the frame display
            self.current_frame = 0  # type: int
            self.current_frame_data = self.tdata[self.current_frame]  # type: np.ndarray
            initialize_vmin = round(self.current_frame_data.min(), 8)
            initialize_vmax = round(self.current_frame_data.max(), 8)
            self.frame_img = self.frame_ax.imshow(
                self.current_frame_data, aspect="auto", cmap="plasma", vmin=initialize_vmin, vmax=initialize_vmax
            )
            self.frame_ax.set_title(f"Frame {self.current_frame}")

            # Setup the profile plot
            self.profile_ax.set_xlabel(generate_label(self.domain_unit))
            self.profile_ax.set_ylabel(generate_label(self.data_unit))
            self.profile_ax.grid(True)

            # Add colorbar with formatter to avoid offset
            formatter = ticker.ScalarFormatter(useMathText=False, useOffset=False)
            self.colorbar = plt.colorbar(self.frame_img, ax=self.frame_ax, format=formatter)

            # 3.) Setup the interactive elements
            # Setup the slider
            slider_ax = plt.axes((0.2, 0.02, 0.6, 0.03))
            self.frame_slider = Slider(
                ax=slider_ax, label="Frame", valmin=0, valmax=self.tdata.shape[0] - 1, valinit=0, valstep=1
            )

            # Setup the clear button
            clear_ax = plt.axes((0.85, 0.02, 0.1, 0.03))
            self.clear_button = Button(clear_ax, "Clear Points")

            # Create checkbox for annotation toggle
            check_ax = plt.axes((0.85, 0.07, 0.1, 0.03))  # Position below clear button
            self.annotation_toggle = CheckButtons(
                check_ax,
                ["Show Value"],
                [True],  # Initially checked
            )

            # 4.) Initialize state variables
            # Store selected points and their profiles
            self.selected_points: list[tuple[int, int]] = []
            self.point_markers: list[Line2D] = []
            self.profile_lines: list[Line2D] = []
            self.colors = ["red", "blue", "green", "purple"]  # Colors for up to 4 points
            self._last_hover_pixel: tuple[int, int] | None = None
            self._closed = False

            # Initialize annotation box once
            self.cursor_annotation_text = TextArea("", textprops={"color": "white", "backgroundcolor": "black"})
            self.cursor_annotation_box = AnnotationBbox(
                self.cursor_annotation_text,
                (0, 0),  # Initial position
                xybox=(10, 10),
                boxcoords="offset points",
                frameon=False,
            )
            self.cursor_annotation_box.set_visible(False)  # Hide initially
            self.frame_ax.add_artist(self.cursor_annotation_box)

            # 5.) Connect events
            self._slider_cid = self.frame_slider.on_changed(self.update_frame)
            self._clear_btn_cid = self.clear_button.on_clicked(self.clear_points)
            self._annotation_cid = self.annotation_toggle.on_clicked(self.toggle_annotation)
            self._canvas_connection_ids = [
                self.fig.canvas.mpl_connect("button_press_event", self.on_click),
                self.fig.canvas.mpl_connect("motion_notify_event", self.on_mouse_move),
                self.fig.canvas.mpl_connect("close_event", self.on_close),
            ]

            # 6.) Initialize blitting for faster rendering (if possible)
            self.fig.canvas.draw_idle()

        @property
        def closed(self) -> bool:
            """Whether the interactive analyzer has been closed and cleaned up."""
            return self._closed

        def close(self, close_figure: bool = False):
            """Disconnect callbacks and release analyzer resources."""
            if self._closed:
                return

            self._closed = True

            for connection_id in self._canvas_connection_ids:
                self.fig.canvas.mpl_disconnect(connection_id)

            self.frame_slider.disconnect(self._slider_cid)
            self.clear_button.disconnect(self._clear_btn_cid)
            self.annotation_toggle.disconnect(self._annotation_cid)

            self.container.release_interactive_analyzer(self)

            if close_figure and plt.fignum_exists(self.fig.number):
                plt.close(self.fig)

        def on_close(self, event):  # pylint: disable=unused-argument
            """Handle figure close event by disconnecting callbacks."""
            self.close(close_figure=False)

        def toggle_annotation(self, event):  # pylint: disable=unused-argument
            """Toggle cursor annotation on/off."""
            # Hide annotation if disabled
            if not self.annotation_toggle.get_status()[0]:
                self.cursor_annotation_box.set_visible(False)
                self._last_hover_pixel = None
                self.fig.canvas.draw_idle()

        def on_mouse_move(self, event):
            """Update annotation when mouse moves over the image."""
            # Check if annotation is enabled
            if not self.annotation_toggle.get_status()[0]:
                return

            if event.inaxes != self.frame_ax:
                if self.cursor_annotation_box.get_visible():
                    self.cursor_annotation_box.set_visible(False)
                    self._last_hover_pixel = None
                    self.fig.canvas.draw_idle()
                return

            if event.xdata is None or event.ydata is None:
                return

            # Get mouse coordinates
            x, y = round(event.xdata), round(event.ydata)

            if 0 <= y < self.current_frame_data.shape[0] and 0 <= x < self.current_frame_data.shape[1]:
                if self._last_hover_pixel == (x, y) and self.cursor_annotation_box.get_visible():
                    return

                # Get current value
                val = self.current_frame_data[y, x]

                # Update annotation
                self.cursor_annotation_box.xy = (x, y)
                self.cursor_annotation_text.set_text(f"({x}, {y})\n{val:.5f}")
                self.cursor_annotation_box.set_visible(True)
                self._last_hover_pixel = (x, y)

                self.fig.canvas.draw_idle()
            elif self.cursor_annotation_box.get_visible():
                self.cursor_annotation_box.set_visible(False)
                self._last_hover_pixel = None
                self.fig.canvas.draw_idle()

        def update_frame(self, frame_idx: float):
            """Update the displayed frame."""
            new_frame = int(frame_idx)
            if new_frame == self.current_frame:
                return

            # Extract frame data
            self.current_frame = new_frame
            self.current_frame_data = self.tdata[new_frame]

            # Invalidate hover cache because pixel values changed with the frame
            self._last_hover_pixel = None
            self.cursor_annotation_box.set_visible(False)

            # Update image data and title
            self.frame_img.set_data(self.current_frame_data)
            self.frame_ax.set_title(f"Frame {self.current_frame}")

            # Get the colorbar limits according to min/max of the current frame
            vmin = round(self.current_frame_data.min(), 8)
            vmax = round(self.current_frame_data.max(), 8)

            old_min, old_max = self.frame_img.get_clim()
            if not np.isclose(old_min, vmin, atol=1e-8, rtol=0.0) or not np.isclose(old_max, vmax, atol=1e-8, rtol=0.0):
                self.frame_img.set_clim(vmin=vmin, vmax=vmax)

            # Redraw
            self.fig.canvas.draw_idle()

        def on_click(self, event):
            """Handle click events on the frame plot."""
            if event.inaxes != self.frame_ax:
                return

            if len(self.selected_points) >= 4:
                print("Maximum number of points (4) reached. Clear points to add more.")
                return

            # Check if click is within frame boundaries
            x, y = int(event.xdata), int(event.ydata)
            if not 0 <= y < self.current_frame_data.shape[0] or not 0 <= x < self.current_frame_data.shape[1]:
                return

            # Add point and plot profile
            color = self.colors[len(self.selected_points)]
            self.selected_points.append((x, y))

            # Plot point on frame
            marker = self.frame_ax.plot(x, y, "x", color=color, markersize=10)[0]
            self.point_markers.append(marker)

            # Plot temperature profile
            profile = self.tdata[:, y, x]
            line = self.profile_ax.plot(self.domain_values, profile, color=color, label=f"Point ({x}, {y})")[0]
            self.profile_lines.append(line)
            self.profile_ax.legend()

            self.fig.canvas.draw_idle()

        def clear_points(self, event):  # pylint: disable=unused-argument
            """Clear all selected points and profiles."""
            self.selected_points.clear()
            self.profile_ax.clear()

            for marker in self.point_markers:
                marker.remove()
            self.point_markers.clear()
            self.profile_lines.clear()

            # Reset profile plot
            self.profile_ax.set_xlabel(generate_label(self.domain_unit))
            self.profile_ax.set_ylabel(generate_label(self.data_unit))
            self.profile_ax.grid(True)

            # Redraw frame without points
            self.frame_img.set_data(self.current_frame_data)

            self.fig.canvas.draw_idle()

    def show_frame(self, frame_number: int, option: str = "", cmap: str = "plasma"):
        """Visualize a specific frame from the dataset with optional ground truth visualization and color mapping.

        Args:
            frame_number (int): The frame number to visualize.
            option (str): The visualization option to apply.
                Options are "ShowGroundTruth", "OverlayGroundTruth", or an empty string.
            cmap (str): The color map to use for the visualization. Defaults to 'plasma'.
        """
        # Clear current figure
        plt.clf()

        # Extract the data from the container
        data = self.get_dataset("/Data/Tdata")
        groundtruth = self.get_dataset("/GroundTruth/DefectMask")

        # Get the frame to show
        data_to_show = data[:, :, frame_number]

        # Show the frame with the selected option
        match option:
            case "ShowGroundTruth":
                plt.subplot(1, 2, 1)
                image = plt.imshow(data_to_show, aspect="auto", cmap=cmap)
                plt.title(f"Frame Number: {frame_number}")

                plt.subplot(1, 2, 2)
                plt.imshow(groundtruth, aspect="auto")
                plt.title("Ground Truth")

            case "OverlayGroundTruth":
                image = plt.imshow(data_to_show, aspect="auto", cmap=cmap)  # Display the original data
                plt.title(f"Frame Number: {frame_number}")

                if groundtruth is not None:
                    # Prepare the overlay
                    binary_gt = groundtruth > 0  # Create a binary mask of the ground truth
                    rows, cols = groundtruth.shape
                    gt_overlay = torch.zeros((rows, cols, 3))  # Initialize an all-zero RGB image for the overlay
                    gt_overlay[:, :, 1] = binary_gt  # Apply green in the binary mask areas

                    plt.imshow(gt_overlay, alpha=0.5)  # Display overlay with transparency

            # Default case, just show the frame data
            case _:
                image = plt.imshow(data_to_show, aspect="auto", cmap=cmap)
                plt.title(f"Frame Number: {frame_number}")

        # Custom formatter for the colorbar to ensure that the colorbar ticks are displayed without offset
        formatter = ticker.ScalarFormatter(useMathText=False, useOffset=False)

        # Show the plot
        plt.colorbar(image, format=formatter)
        plt.show()

    def show_pixel_profile(self, pixel_pos_x: int, pixel_pos_y: int):
        """Plot the profile of a specific pixel across the dataset's domain values with an option for data adjustment.

        The X-axis of the plot is labeled according to the domaintype attribute, reflecting the dataset's domain
        (e.g., time, frequency). The Y-axis is generically labeled as 'Temperature in K'.

        Args:
            pixel_pos_x (int): The X-coordinate (column index) of the pixel.
                Must be within the dataset's second dimension range.
            pixel_pos_y (int): The Y-coordinate (row index) of the pixel.
                Must be within the dataset's first dimension range.
        """
        # Clear the current figure
        plt.clf()

        # Extract the data from the container
        data = self.get_dataset("/Data/Tdata")
        domainvalues = self.get_dataset("/MetaData/DomainValues")
        data_unit = self.get_unit("/Data/Tdata")
        domain_unit = self.get_unit("/MetaData/DomainValues")

        # Validate pixel positions to be within the data dimensions
        if pixel_pos_x < 0 or pixel_pos_y < 0 or pixel_pos_x >= data.shape[0] or pixel_pos_y >= data.shape[1]:
            raise ValueError("Pixel positions must be within the range of data dimensions.")

        # Extract temperature profile of the pixel
        temperature_profile = data[pixel_pos_y, pixel_pos_x, :]

        # Plot the temperature profile
        plt.plot(domainvalues, temperature_profile)
        plt.title(f"Profile of Pixel: {pixel_pos_x},{pixel_pos_y}")
        plt.xlabel(generate_label(domain_unit))
        plt.ylabel(generate_label(data_unit))
        plt.show()

    def release_interactive_analyzer(self, analyzer: "VisualizationOps.InteractiveAnalyzer") -> None:
        """Release the stored interactive analyzer if it matches the provided instance."""
        if self._interactive_analyzer is analyzer:
            self._interactive_analyzer = None

    def analyse_interactive(self):
        """Launch interactive analysis session for thermographic data visualization."""
        if self._interactive_analyzer is not None and not self._interactive_analyzer.closed:
            self._interactive_analyzer.close(close_figure=True)
        self._interactive_analyzer = self.InteractiveAnalyzer(self)
        plt.show()



import logging
import warnings
from functools import lru_cache
from importlib.metadata import entry_points

from .base_parser import BaseParser
from .edevis_parser import EdevisParser
from .hdf5_parser import HDF5Parser
from .simulation_parser import SimulationParser

logger = logging.getLogger(__name__)


def _load_parser_plugins() -> tuple[type[BaseParser], ...]:
    """Load parser plugins via entry points."""
    plugins = []
    for ep in entry_points(group="pythermondt.parsers"):
        try:
            parser_cls = ep.load()
            plugins.append(parser_cls)
            logger.debug("Loaded parser plugin '%s' from '%s'", ep.name, ep.module)
        except Exception as e:  # pylint: disable=broad-except
            warnings.warn(f"Failed to load parser plugin '{ep.name}': {e}", stacklevel=2)
    return tuple(plugins)


@lru_cache(maxsize=1)
def _get_registry() -> tuple[type[BaseParser], ...]:
    """Lazily build and cache the parser registry on first use.

    Constructs the registry of built-in and plugin parsers only when first called.
    The result is cached so subsequent calls return the same registry instance. A tuple of parser classes is returned,
    ensuring immutability and thread safety, exactly as the previous module level CONSTANT list did.
    """
    builtins = (HDF5Parser, SimulationParser, EdevisParser)
    plugins = _load_parser_plugins()
    return builtins + plugins


def find_parser_for_extension(extension: str) -> type[BaseParser] | None:
    """Find a parser that supports the given file extension.

    Args:
        extension: File extension (with or without leading dot)

    Returns:
        Parser class that supports the extension, or None if not found
    """
    # Normalize extension to include leading dot
    normalized_ext = extension if extension.startswith(".") else f".{extension}"

    # Find first parser supporting this extension
    for parser_cls in _get_registry():
        if normalized_ext in parser_cls.supported_extensions:
            return parser_cls

    return None


def get_all_supported_extensions() -> set[str]:
    """Get all supported extensions of all registered parsers.

    Returns:
        Set of all supported extensions
    """
    return {ext for parser_cls in _get_registry() for ext in parser_cls.supported_extensions}


def get_all_parsers() -> tuple[type[BaseParser], ...]:
    """Get all registered parsers.

    Returns:
        Tuple of all parser classes
    """
    return _get_registry()


__all__ = [
    "BaseParser",
    "EdevisParser",
    "HDF5Parser",
    "SimulationParser",
    "find_parser_for_extension",
    "get_all_parsers",
    "get_all_supported_extensions",
]



from azure.core.credentials import TokenCredential

from ..data import DataContainer
from ..data.datacontainer.serialization_ops import CompressionType
from ..io import AzureBlobBackend, IOPathWrapper
from .base_writer import BaseWriter


class AzureBlobWriter(BaseWriter):
    def __init__(
        self,
        account_url: str,
        container_name: str,
        prefix: str,
        connection_string: str | None = None,
        credential: TokenCredential | None = None,
    ):
        """Initialize writer for Azure Blob Storage.

        Uses DefaultAzureCredential by default (managed identity, az login, env vars, etc.).
        Explicit connection_string or credential override the default behavior.

        Args:
            account_url (str): Storage account URL (https://<account>.blob.core.windows.net)
            container_name (str): Azure blob container name
            prefix (str): Prefix (folder path) within container
            connection_string (str | None): Optional connection string (overrides default auth)
            credential (TokenCredential | None): Optional Azure TokenCredential (overrides default auth)
        """
        super().__init__()
        self.__account_url = account_url
        self.__container_name = container_name
        self.__prefix = prefix
        self.__connection_string = connection_string
        self.__credential = credential

    def _create_backend(self) -> AzureBlobBackend:
        """Create a new AzureBlobBackend instance."""
        # pylint: disable=duplicate-code
        return AzureBlobBackend(
            account_url=self.__account_url,
            container_name=self.__container_name,
            prefix=self.__prefix,
            connection_string=self.__connection_string,
            credential=self.__credential,
        )
        # pylint: enable=duplicate-code

    def write(
        self,
        container: DataContainer,
        file_name: str,
        compression: CompressionType = "lzf",
        compression_opts: int | None = 4,
    ):
        """Write DataContainer to Azure Blob Storage.

        Args:
            container (DataContainer): The data container to write
            file_name (str): Name of the file to write (will append .hdf5 if missing)
            compression (CompressionType): Compression algorithm to use. Default: "lzf"
            compression_opts (int | None): Compression level (algorithm-dependent). Default: 4
        """
        # Append file extension if not present
        if not file_name.endswith(".hdf5"):
            file_name += ".hdf5"

        # Writer constructs full blob name with prefix
        full_blob_name = f"{self.__prefix}/{file_name}" if self.__prefix else file_name

        # Backend handles everything - just pass the blob name
        data = container.serialize_to_hdf5(compression, compression_opts)
        self.backend.write_file(IOPathWrapper(data), full_blob_name)



import logging
from io import BytesIO
from typing import IO, cast

from azure.core.credentials import TokenCredential
from azure.core.exceptions import AzureError, ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from tqdm.auto import tqdm

from ..utils import IOPathWrapper
from .base_backend import BaseBackend
from .progress import TqdmCallback, get_tqdm_default_kwargs

logger = logging.getLogger(__name__)


class AzureBlobBackend(BaseBackend):
    def __init__(
        self,
        account_url: str,
        container_name: str,
        prefix: str = "",
        connection_string: str | None = None,
        credential: TokenCredential | None = None,
    ) -> None:
        """Initialize Azure Blob Storage backend.

        Uses DefaultAzureCredential by default (managed identity, az login, env vars, etc.). Providing an explicit
        connection_string or credential overrides the default behavior.

        Authentication priority:
            1. DefaultAzureCredential (default - auto-discovers credentials)
            2. connection_string (if provided overrides default behavior)
            3. credential (if provided overrides default behavior)

        Args:
            account_url (str): Storage account URL (https://<account>.blob.core.windows.net)
            container_name (str): Container name
            prefix (str, optional): Prefix within the specified container
            connection_string (str, optional): Connection string (optional, for dev/researchers)
            credential (TokenCredential, optional): Azure TokenCredential (optional, defaults to DefaultAzureCredential)
        """
        if connection_string:
            self.__client = BlobServiceClient.from_connection_string(connection_string)
            logger.debug("Client initialized using connection string.")
        else:
            if credential is None:
                credential = DefaultAzureCredential()
                logger.debug("Using DefaultAzureCredential for authentication.")
            self.__client = BlobServiceClient(account_url, credential=credential)

        self.__container_name = container_name
        self.__prefix = prefix

        logger.debug("AzureBlobBackend(container=%s, prefix=%s) initialized.", container_name, prefix)

    @property
    def remote_source(self) -> bool:
        return True

    @property
    def scheme(self) -> str:
        return "az"

    @property
    def container_name(self) -> str:
        return self.__container_name

    @property
    def prefix(self) -> str:
        return self.__prefix

    def read_file(self, file_path: str) -> IOPathWrapper:
        """Read a file from Azure Blob Storage.

        Args:
            file_path (str): Path to file, either full Azure URI or blob name within container

        Returns:
            IOPathWrapper: File contents

        Raises:
            FileNotFoundError: If file doesn't exist
        """
        container, blob_name = self._parse_input(file_path)

        try:
            blob_client = self.__client.get_blob_client(container=container, blob=blob_name)

            # Get blob properties for progress bar
            properties = blob_client.get_blob_properties()
            file_size = properties.size

            data = BytesIO()
            with TqdmCallback(total=file_size, desc=f"Downloading {blob_name}") as pbar:
                stream = blob_client.download_blob()
                for chunk in stream.chunks():
                    data.write(chunk)
                    pbar.callback(len(chunk))

            data.seek(0)
            return IOPathWrapper(data)

        except ResourceNotFoundError as e:
            raise FileNotFoundError(f"File not found: {file_path}") from e

    def write_file(self, data: IOPathWrapper, file_path: str) -> None:
        """Write file to Azure Blob Storage.

        Args:
            data (IOPathWrapper): File data to write
            file_path (str): Destination path

        Raises:
            RuntimeError: If upload fails
        """
        container, blob_name = self._parse_input(file_path)

        # Reset and get size
        file_size = data.file_obj.getbuffer().nbytes

        try:
            blob_client = self.__client.get_blob_client(container=container, blob=blob_name)

            # Wrap file object to track read progress
            data.file_obj.seek(0)
            with tqdm.wrapattr(
                data.file_obj, "read", desc=f"Uploading {blob_name}", **get_tqdm_default_kwargs(file_size=file_size)
            ) as wrapped_file:
                blob_client.upload_blob(cast(IO[bytes], wrapped_file), overwrite=True)

        except AzureError as e:
            raise RuntimeError(f"Failed to upload blob: {e}") from e

    def exists(self, file_path: str) -> bool:
        """Check if a file exists in Azure Blob Storage.

        Args:
            file_path (str): Path to check

        Returns:
            bool: True if file exists
        """
        container, blob_name = self._parse_input(file_path)

        try:
            blob_client = self.__client.get_blob_client(container=container, blob=blob_name)
            blob_client.get_blob_properties()
            return True
        except ResourceNotFoundError:
            return False

    def close(self) -> None:
        """Close connections.

        For Azure Blob Storage, closes the underlying BlobServiceClient.
        """
        self.__client.close()

    def get_file_list(self, extensions: tuple[str, ...] | None = None, num_files: int | None = None) -> list[str]:
        """Get list of files from Azure Blob Storage with optional filtering.

        Args:
            extensions (tuple[str, ...], optional): Filter by file extensions
            num_files (int, optional): Limit number of files returned

        Returns:
            list[str]: List of file paths
        """
        blobs = []
        container_client = self.__client.get_container_client(self.__container_name)

        for blob in container_client.list_blobs(name_starts_with=self.prefix):
            if blob.name.endswith("/"):
                continue

            # Add full Azure path using _to_url
            blobs.append(self._to_url(self.__container_name, blob.name))

        if extensions:
            blobs = [b for b in blobs if any(b.lower().endswith(ext.lower()) for ext in extensions)]

        blobs.sort()

        if num_files is not None:
            blobs = blobs[:num_files]

        return blobs

    def get_file_size(self, file_path: str) -> int:
        """Get the size of the file in Azure Blob Storage in bytes.

        Args:
            file_path (str): Path to file

        Returns:
            int: File size in bytes

        Raises:
            FileNotFoundError: If file doesn't exist
        """
        container, blob_name = self._parse_input(file_path)

        try:
            blob_client = self.__client.get_blob_client(container=container, blob=blob_name)

            properties = blob_client.get_blob_properties()
            return properties.size
        except ResourceNotFoundError as e:
            raise FileNotFoundError(f"File not found: {file_path}") from e

    def get_file_identity(self, file_path: str) -> str:
        """Get blob identity for change detection.

        Args:
            file_path (str): Path to file.

        Returns:
            str: Blob ETag.

        Raises:
            FileNotFoundError: If file doesn't exist.
        """
        container, blob_name = self._parse_input(file_path)

        try:
            blob_client = self.__client.get_blob_client(container=container, blob=blob_name)
            properties = blob_client.get_blob_properties()
            etag = properties.etag
            if etag is None:
                raise RuntimeError(f"ETag unavailable for file: {file_path}")
            return str(etag)
        except ResourceNotFoundError as e:
            raise FileNotFoundError(f"File not found: {file_path}") from e

    def download_file(self, source_path: str, destination_path: str) -> None:
        """Download a file from Azure Blob Storage to local filesystem.

        Args:
            source_path (str): Source Azure path
            destination_path (str): Destination local path
        """
        container, blob_name = self._parse_input(source_path)

        blob_client = self.__client.get_blob_client(container=container, blob=blob_name)

        # Get file size for progress
        properties = blob_client.get_blob_properties()
        file_size = properties.size

        with open(destination_path, "wb") as local_file:
            with TqdmCallback(total=file_size, desc=f"Downloading {blob_name}") as progress:
                stream = blob_client.download_blob()
                for chunk in stream.chunks():
                    local_file.write(chunk)
                    progress.callback(len(chunk))

    def _parse_input(self, file_path: str) -> tuple[str, str]:
        """Convert Azure URI to (container, blob_name) tuple.

        Args:
            file_path: Either "az://container/blob" or just "blob"

        Returns:
            tuple[str, str]: (container, blob_name)
        """
        if file_path.startswith("az://"):
            # az://container/blob/path -> container="container", blob_name="blob/path"
            path = file_path[5:]  # Remove "az://"
            parts = path.split("/", 1)
            container = parts[0]
            blob_name = parts[1] if len(parts) > 1 else ""
            return container, blob_name

        # Not a URI - treat as blob name within default container
        # No prefix prepending - user must provide full blob name
        return self.__container_name, file_path

    def _to_url(self, container: str, blob_name: str) -> str:
        """Convert (container, blob_name) to Azure URI.

        Args:
            container: Azure container name
            blob_name: Blob name/key

        Returns:
            str: Azure URI like "az://container/blob"
        """
        return f"az://{container}/{blob_name}"



import logging
import os
from glob import glob
from urllib.parse import urlparse
from urllib.request import pathname2url, url2pathname

from ..utils import IOPathWrapper
from .base_backend import BaseBackend

logger = logging.getLogger(__name__)


class LocalBackend(BaseBackend):
    def __init__(self, pattern: str, recursive: bool = False) -> None:
        """Initialize an instance of the LocalBackend class.

        This class is used to read data from local files, or directories, using the standard Python file I/O operations.

        Args:
            pattern (str): The pattern that will be used to match files to read. This can either be a file path, a
                directory path, or a glob pattern.
            recursive (bool): If True, the pattern will be applied recursively to all subdirectories. This will only
                be effective if the pattern is a directory path or a glob pattern. Defaults to False.
        """
        # Validate the pattern
        if not isinstance(pattern, str):
            raise ValueError(f"Invalid pattern type: {type(pattern)}. Must be a string.")
        if pattern == "":
            raise ValueError(f"Invalid pattern: {pattern!r}. Must be a non-empty string.")

        # Initialize the pattern either as url or path
        parsed_input = self._parse_input(pattern)

        # Determine the type of the source based on the provided pattern
        self.__source_type = None
        if os.path.isfile(parsed_input):
            self.__source_type = "file"
        elif os.path.isdir(parsed_input):
            self.__source_type = "directory"
        else:
            self.__source_type = "glob"
        logger.debug("Source type determined as: %s", self.__source_type)

        # Internal state
        self.__pattern_str = parsed_input
        self.__recursive = recursive
        logger.debug("LocalBackend(pattern=%s, recursive=%s) initialized.", pattern, recursive)

    @property
    def remote_source(self) -> bool:
        return False

    @property
    def scheme(self) -> str:
        return "file"

    @property
    def pattern(self) -> str:
        return self.__pattern_str

    def read_file(self, file_path: str) -> IOPathWrapper:
        path = self._parse_input(file_path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {file_path}")
        return IOPathWrapper(path)

    def write_file(self, data: IOPathWrapper, file_path: str) -> None:
        path = self._parse_input(file_path)
        with open(path, "wb") as file:
            file.write(data.file_obj.read())

    def exists(self, file_path: str) -> bool:
        return os.path.exists(self._parse_input(file_path))

    def close(self) -> None:
        # Nothing to close for local files
        pass

    def get_file_list(self, extensions: tuple[str, ...] | None = None, num_files: int | None = None) -> list[str]:
        # Handle different pattern types
        all_files = []
        match self.__source_type:
            case "file":
                all_files = [self.pattern]
            case "directory":
                if self.__recursive:
                    all_files = [os.path.join(root, name) for root, _, names in os.walk(self.pattern) for name in names]
                else:
                    all_files = [f.path for f in os.scandir(self.pattern) if f.is_file()]
            case "glob":
                all_files = glob(self.pattern, recursive=self.__recursive)

        # Filter by extension if provided
        if extensions:
            all_files = [f for f in all_files if any(f.endswith(ext) for ext in extensions)]

        # Sort for deterministic behavior
        all_files.sort()

        # Limit number of results if specified
        if num_files is not None:
            all_files = all_files[:num_files]

        # Convert to absolute paths and normalize before returning
        all_files = [self._to_url(os.path.normpath(os.path.abspath(f))) for f in all_files]

        return all_files

    def get_file_size(self, file_path: str) -> int:
        path = self._parse_input(file_path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {file_path}")
        if os.path.isdir(path):
            raise IsADirectoryError(f"Path is a directory, not a file: {path}")
        return os.path.getsize(path)

    def get_file_identity(self, file_path: str) -> str:
        """Return a low-overhead identity string for local files.

        The identity is based on filesystem metadata and intended for change
        detection, not cryptographic integrity verification.
        """
        path = self._parse_input(file_path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"File not found: {file_path}")
        if os.path.isdir(path):
            raise IsADirectoryError(f"Path is a directory, not a file: {path}")

        stat_result = os.stat(path)
        device = getattr(stat_result, "st_dev", 0)
        inode = getattr(stat_result, "st_ino", 0)

        return f"{device}:{inode}:{stat_result.st_size}:{stat_result.st_mtime_ns}"

    def download_file(self, source_path: str, destination_path: str) -> None:
        raise NotImplementedError("Direct download is not supported for local files.")

    def _parse_input(self, input_path: str) -> str:
        parsed = urlparse(input_path)
        return url2pathname(parsed.path) if parsed.scheme == self.scheme else input_path

    def _to_url(self, internal_path: str) -> str:
        url_path = pathname2url(internal_path)
        return f"file:{url_path}" if url_path.startswith("///") else f"file://{url_path}"



import logging
from io import BytesIO
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError

from ..utils import IOPathWrapper
from .base_backend import BaseBackend
from .progress import TqdmCallback

logger = logging.getLogger(__name__)


class S3Backend(BaseBackend):
    def __init__(self, bucket: str, prefix: str, session: boto3.Session | None = None) -> None:
        # Use default boto3 session if none is provided
        if not session:
            logger.debug("No boto3 session provided, creating default session.")
            session = boto3.Session()

        # Create a new s3 client from the given session
        self.__client = session.client("s3")

        # Write the bucket and prefix to the private attributes
        self.__bucket = bucket
        self.__prefix = prefix
        logger.debug("S3Backend(bucket=%s, prefix=%s) initialized.", bucket, prefix)

    @property
    def remote_source(self) -> bool:
        # This backend is always remote
        return True

    @property
    def scheme(self) -> str:
        return "s3"

    @property
    def bucket(self) -> str:
        return self.__bucket

    @property
    def prefix(self) -> str:
        return self.__prefix

    def read_file(self, file_path: str) -> IOPathWrapper:
        """Read a file from S3.

        Args:
            file_path (str): Path to file, either full S3 URI or key within bucket

        Returns:
            IOPathWrapper: File contents

        Raises:
            FileNotFoundError: If file doesn't exist
        """
        bucket, key = self._parse_input(file_path)
        data = BytesIO()
        with TqdmCallback(total=self.get_file_size(file_path), desc=f"Downloading {key}") as pbar:
            self.__client.download_fileobj(bucket, key, data, Callback=pbar.callback)
        return IOPathWrapper(data)

    def write_file(self, data: IOPathWrapper, file_path: str) -> None:
        """Write file to S3.

        Args:
            data (IOPathWrapper): File data to write
            file_path (str): Destination path
        """
        bucket, key = self._parse_input(file_path)

        # Reset file object position
        data.file_obj.seek(0)

        # Upload to S3 (Always show progress)
        try:
            with TqdmCallback(total=data.file_obj.getbuffer().nbytes, desc=f"Uploading {key}") as pbar:
                self.__client.upload_fileobj(data.file_obj, bucket, key, Callback=pbar.callback)
        except ClientError as e:
            raise RuntimeError(f"Failed to upload file to S3: {e}") from e

    def exists(self, file_path: str) -> bool:
        """Check if a file exists in S3.

        Args:
            file_path (str): Path to check

        Returns:
            bool: True if file exists
        """
        bucket, key = self._parse_input(file_path)

        try:
            self.__client.head_object(Bucket=bucket, Key=key)
            return True
        except ClientError as e:
            if self._is_not_found_error(e):
                return False
            raise

    def close(self) -> None:
        """Close connections.

        For S3, we need to close the underlying boto3 client.
        """
        self.__client.close()

    def get_file_list(self, extensions: tuple[str, ...] | None = None, num_files: int | None = None) -> list[str]:
        """Get list of files from S3 with optional filtering.

        Args:
            extensions (tuple[str, ...], optional): Filter by file extensions
            num_files (int, optional): Limit number of files returned

        Returns:
            list[str]: List of file paths
        """
        # List all objects with the given prefix
        paginator = self.__client.get_paginator("list_objects_v2")

        files = []
        for page in paginator.paginate(Bucket=self.bucket, Prefix=self.prefix):
            if "Contents" in page:
                for obj in page["Contents"]:
                    key = obj["Key"]
                    # Skip directories (keys ending with /)
                    if key.endswith("/"):
                        continue

                    # Add full S3 path
                    files.append(self._to_url(self.bucket, key))

        # Filter by extension if provided
        if extensions:
            files = [f for f in files if any(f.lower().endswith(ext.lower()) for ext in extensions)]

        # Sort for deterministic behavior
        files.sort()

        # Limit results if specified
        if num_files is not None:
            files = files[:num_files]

        return files

    def get_file_size(self, file_path: str) -> int:
        """Return the size of the file on s3 bucket in bytes."""
        bucket, key = self._parse_input(file_path)

        try:
            response = self.__client.head_object(Bucket=bucket, Key=key)
            return response["ContentLength"]
        except ClientError as e:
            if self._is_not_found_error(e):
                raise FileNotFoundError(f"File not found: {file_path}") from e
            raise

    def get_file_identity(self, file_path: str) -> str:
        """Return object identity (ETag) for a file on S3.

        Args:
            file_path (str): Path to file on S3.

        Returns:
            str: ETag value.

        Raises:
            FileNotFoundError: If file doesn't exist.
        """
        bucket, key = self._parse_input(file_path)

        try:
            response = self.__client.head_object(Bucket=bucket, Key=key)
            etag = response.get("ETag")
            if etag is None:
                raise RuntimeError(f"ETag unavailable for file: {file_path}")
            return etag
        except ClientError as e:
            if self._is_not_found_error(e):
                raise FileNotFoundError(f"File not found: {file_path}") from e
            raise

    def download_file(self, source_path: str, destination_path: str) -> None:
        """Download a file from S3 to local filesystem.

        Args:
            source_path (str): Source S3 path
            destination_path (str): Destination local path
        """
        bucket, key = self._parse_input(source_path)

        # Download the file
        with TqdmCallback(total=self.get_file_size(source_path), desc=f"Downloading {key}") as progress:
            self.__client.download_file(bucket, key, destination_path, Callback=progress.callback)

    def _parse_input(self, file_path: str) -> tuple[str, str]:
        """Convert S3 URI to (bucket, key) tuple.

        Args:
            file_path: Either "s3://bucket/key" or just "key"

        Returns:
            tuple[str, str]: (bucket, key)
        """
        parsed = urlparse(file_path)
        if parsed.scheme == "s3":
            # s3://bucket/key/path -> bucket="bucket", key="key/path"
            bucket = parsed.netloc
            key = parsed.path.lstrip("/")  # Remove leading slash
            return bucket, key

        # Not a URI - treat as absolute key within default bucket
        # No prefix prepending - user must provide full key
        return self.bucket, file_path

    def _to_url(self, bucket: str, key: str) -> str:
        """Convert (bucket, key) to S3 URI.

        Args:
            bucket: S3 bucket name
            key: Object key

        Returns:
            str: S3 URI like "s3://bucket/key"
        """
        return f"s3://{bucket}/{key}"

    def _is_not_found_error(self, e: ClientError) -> bool:
        """Check if ClientError indicates file not found.

        AWS S3 returns different error codes for "not found":
        - '404' - from head_object() operation
        - 'NoSuchKey' - from get_object() and other operations
        - 'NoSuchBucket' - bucket doesn't exist
        - '403' - can indicate missing file if user lacks s3:ListBucket permission

        Args:
            e: ClientError from boto3

        Returns:
            bool: True if this is a not-found error
        """
        error_code = e.response.get("Error", {}).get("Code", "")
        return error_code in ("404", "403", "NoSuchKey", "NoSuchBucket")



import re
from typing import TypeGuard

from .node import BaseNode, DataNode, GroupNode, NodeType, RootNode


def is_rootnode(node: BaseNode) -> "TypeGuard[RootNode]":
    """Checks if the given node is a RootNode.

    Args:
        node (BaseNode): The node to check.

    Returns:
        bool: True if the node is a RootNode, False otherwise.
    """
    return isinstance(node, RootNode) and node.type == NodeType.ROOT


def is_groupnode(node: BaseNode) -> "TypeGuard[GroupNode]":
    """Checks if the given node is a GroupNode.

    Args:
        node (BaseNode): The node to check.

    Returns:
        bool: True if the node is a GroupNode, False otherwise.
    """
    return isinstance(node, GroupNode) and node.type == NodeType.GROUP


def is_datanode(node: BaseNode) -> "TypeGuard[DataNode]":
    """Checks if the given node is a DataNode.

    Args:
        node (BaseNode): The node to check.

    Returns:
        bool: True if the node is a DataNode, False otherwise.
    """
    return isinstance(node, DataNode) and node.type == NodeType.DATASET


def validate_path(path: str, name: str = "") -> str:
    """Validates and normalizes the given HDF5 path.

    Args:
        path (str): The path to validate and normalize.
        name (str, optional): The name of the group or dataset to add to the path. Defaults to "".

    Returns:
        A normalized valid HDF5 path.

    Raises:
        ValueError: If the path is not valid.
    """
    # Add name to the path
    if name and path[-1] != "/":
        path = path + "/" + name
    else:
        path = path + name

    # Normalize: strip leading/trailing whitespace and ensure starting with a slash
    normalized_path = path.strip()
    if not normalized_path.startswith("/"):
        normalized_path = "/" + normalized_path

    # Check for double slashes or trailing slash ==> remove them
    if "//" in normalized_path:
        normalized_path = re.sub(r"/+", "/", normalized_path)

    if normalized_path != "/" and normalized_path.endswith("/"):
        normalized_path = normalized_path[:-1]

    # Validate using a regex pattern
    pattern = r"^/[a-zA-Z0-9_/-]+$"
    if not re.match(pattern, normalized_path):
        raise ValueError(f"Invalid path: {normalized_path} contains invalid characters.")

    return normalized_path


def validate_paths(paths: list[str]) -> tuple[str, ...]:
    """Validates and normalizes the given list of HDF5 paths.

    Args:
        paths (List[str]): The list of paths to validate and normalize.

    Returns:
        A list of normalized valid HDF5 paths.

    Raises:
        ValueError: If any of the paths is not valid.
    """
    return tuple(validate_path(path) for path in paths)


def split_path(path: str) -> tuple[str, str]:
    """Splits the given HDF5 path into the parent path and the name of the group or dataset.

    Args:
        path (str): The path to split.

    Returns:
        parent (str): The parent path.
        child (str): The name of the group or dataset.
    """
    parent, child = path.rsplit("/", 1)
    if path.count("/") == 1:
        return "/", child
    return parent, child


def generate_key(path: str, name: str) -> tuple[str, str, str]:
    """Generates a key for the given path and name.

    First the path is validated and normalized, then it is split into the parent path and the name of the
    group or dataset.

    Args:
        path (str): The path to generate the key for.
        name (str): The name of the group or dataset.

    Returns:
        key (str): The generated key to be saved in the dictionary.
        parent (str): The parent path.
        child (str): The name of the group or dataset.
    """
    key = validate_path(path, name)
    head, tail = split_path(key)
    return key, head, tail


def format_bytes(size_bytes: int) -> str:
    """Format bytes into human-readable string with appropriate unit."""
    size = float(size_bytes)  # Ensure size_bytes is a float for division
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024:
            return f"{size:.2f} {unit}"
        size /= 1024
    return f"{size:.2f} PB"  # Fallback for very large sizes



from ..data import units
from .datacontainer import DataContainer


class ThermoContainer(DataContainer):
    def __init__(self):
        """Initializes a DataContainer with a predefined structure for thermographic data."""
        super().__init__()

        # Set initial groups
        self.add_group("/", "Data")
        self.add_group("/", "GroundTruth")
        self.add_group("/", "MetaData")

        # Set initial datasets
        self.add_dataset("/Data", "Tdata")
        self.add_dataset("/GroundTruth", "DefectMask")
        self.add_dataset("/MetaData", "LookUpTable")
        self.add_dataset("/MetaData", "ExcitationSignal")
        self.add_dataset("/MetaData", "DomainValues")

        # Add units to the initial datasets
        self.add_attributes(
            "/Data/Tdata", Unit=units.arbitrary
        )  # Tdata has a arbitrary unit because it is raw data and the LUT has not been applied yet
        self.add_attributes("/MetaData/DomainValues", Unit=units.second)
        self.add_attributes("/MetaData/LookUpTable", Unit=units.kelvin)
        self.add_attributes("/MetaData/ExcitationSignal", Unit=units.dimensionless)
        self.add_attributes("/GroundTruth/DefectMask", Unit=units.dimensionless)



from collections.abc import ItemsView, KeysView, ValuesView
from sys import getsizeof
from typing import TypeVar, overload

from .node import DataNode, GroupNode, RootNode
from .utils import format_bytes, split_path

# Add typevars to make type hinting more readable
NodeTypes = RootNode | GroupNode | DataNode
T = TypeVar("T", bound=NodeTypes)


# Custom class to access nodes in the DataContainer ==> Needed to make the node property indexable and callable
class NodeAccessor:
    def __init__(self):
        self.__nodes: dict[str, NodeTypes] = {}

    # Functions to get views of the __nodes dictionary
    def keys(self) -> KeysView[str]:
        return self.__nodes.keys()

    def values(self) -> ValuesView[NodeTypes]:
        return self.__nodes.values()

    def items(self) -> ItemsView[str, NodeTypes]:
        return self.__nodes.items()

    # Private methods to get, set and delete nodes
    @overload
    def __get_node(self, key: str) -> NodeTypes: ...

    @overload
    def __get_node(self, key: str, *node_types: type[T]) -> T: ...

    def __get_node(self, key: str, *node_types: type[T]) -> NodeTypes | T:
        # Check if the path exists
        if key not in self.__nodes:
            raise KeyError(f"Node at path '{key}' does not exist.")

        # Optionally check if the node is of the correct type
        node = self.__nodes[key]
        if node_types and not isinstance(node, node_types):
            raise TypeError(f"Node at path '{key}' is not of type: {', '.join([t.__name__ for t in node_types])}.")
        return node

    def __set_node(self, key: str, value: NodeTypes) -> None:
        # Block any overwrites by default ==> updating nodes is handled using __get_node
        # by overriding .data attribute in Node classes

        # Special handling for root nodes
        if isinstance(value, RootNode):
            if key != "/":
                raise ValueError("RootNode must be placed at the root path '/'.")

            if "/" in self.__nodes:
                raise ValueError(
                    "RootNode already exists in the DataContainer. RootNode must be unique for each DataContainer."
                )

            self.__nodes[key] = value
            return

        if key in self.__nodes:
            raise KeyError(f"Node at path '{key}' already exists. Use a different path or delete the existing node.")

        # Split path in parent and child and check if parent exists
        parent, _ = split_path(key)
        if parent not in self.__nodes:
            if parent == "/" and parent not in self.__nodes:
                raise KeyError(
                    "RootNode does not exist in this container. "
                    "Check container initialization and ensure that a RootNode exists!"
                )
            raise KeyError(f"Parent node at path '{parent}' does not exist.")

        # Also check if parent is a RootNode or GroupNode
        if not isinstance(self.__nodes[parent], (RootNode, GroupNode)):
            raise TypeError(f"Parent node at path '{parent}' must be a RootNode or GroupNode.")

        # Set node at path in dictionary
        self.__nodes[key] = value

    def __delete_node(self, key: str) -> None:
        # Check if the path exists
        if key not in self.__nodes:
            raise KeyError(f"Node at path '{key}' does not exist.")

        # If it's a group, remove all child nodes as well to avoid orphaned nodes
        if isinstance(self.__nodes[key], GroupNode):
            child_keys = [k for k in self.__nodes if k.startswith(key + "/")]
            for child_key in child_keys:
                del self.__nodes[child_key]

        del self.__nodes[key]

    # Overriding the getitem, setitem and delitem methods to access nodes ==> Makes the node property indexable
    def __getitem__(self, key: str) -> NodeTypes:
        return self.__get_node(key)

    def __delitem__(self, key: str) -> None:
        self.__delete_node(key)

    def __setitem__(self, key: str, value: NodeTypes):
        self.__set_node(key, value)

    # Overriding the call method to access nodes ==> Makes the node property callable like a function
    @overload
    def __call__(self, key: str) -> NodeTypes: ...

    @overload
    def __call__(self, key: str, *node_types: type[T]) -> T: ...

    def __call__(self, key: str, *node_types: type[T]) -> NodeTypes | T:
        if not node_types:
            return self.__get_node(key)
        return self.__get_node(key, *node_types)


class DataContainerBase:
    def __init__(self):
        if self.__class__ is DataContainerBase:
            raise TypeError(f"{self.__class__.__name__} is a base class and cannot be instantiated directly.")
        self.__node_accessor = NodeAccessor()

    @property
    def nodes(self) -> NodeAccessor:
        """Property to access nodes in the DataContainer.

        This property is indexable and callable to get and set nodes. The path is checked for existence. Optionally the
        nodes is also checked for the correct type (provided as a function argument) while getting a node. Therefore the
        returned type is ensured to be correct. When setting nodes, the parent path is checked for existence and type
        (either RootNode or GroupNode). It also allows for deleting nodes by using the del keyword. If a group is
        deleted, all child nodes are also deleted to avoid orphaned nodes. Use with caution! Some sanity checks are
        disabled when using the property directly. Only for advanced users!

        Example usage:
        ```python
        from pythermondt.data import DataContainer
        from pythermondt.data.datacontainer.node import GroupNode, DataNode

        # Get nodes
        # Returns the node at path "/Data"
        node = data_container.nodes("/Data")

        # Returns the node at path "/Data". Raises TypeError if the node is not a DataNode
        node = data_container.nodes("/Data", DataNode)

        # Returns the node at path "/Data". Raises TypeError if the node is not a DataNode or GroupNode
        node = data_container.nodes("/Data", DataNode, GroupNode)

        # Set nodes
        data_container.nodes["/Data"] = DataNode("name", data)
        data_container.nodes["/Data"] = GroupNode("name")

        # Delete a node
        del data_container.nodes["/Data"]
        ```
        """
        return self.__node_accessor


# Basic operations for the DataContainer
class BaseOps(DataContainerBase):
    def _path_exists(self, key: str) -> bool:
        """Check if a path exists in the DataContainer.

        Args:
            key (str): The path to check.

        Returns:
            bool: True if the path exists, False otherwise.
        """
        try:
            self.nodes(key)
            return True
        except KeyError:
            return False

    def _is_datanode(self, key: str) -> bool:
        """Check if a DataNode exists at the given path.

        Args:
            key (str): The path to check.

        Returns:
            bool: True if a DataNode exists at the path, False otherwise.
        """
        try:
            self.nodes(key, DataNode)
            return True
        except (KeyError, TypeError):
            return False

    def _is_groupnode(self, key: str) -> bool:
        """Check if a GroupNode exists at the given path.

        Args:
            key (str): The path to check.

        Returns:
            bool: True if a GroupNode exists at the path, False otherwise.
        """
        try:
            self.nodes(key, GroupNode)
            return True
        except (KeyError, TypeError):
            return False

    def _is_rootnode(self, key: str) -> bool:
        """Check if a RootNode exists at the given path.

        Args:
            key (str): The path to check.

        Returns:
            bool: True if the given path is the root path, False otherwise.
        """
        try:
            self.nodes(key, RootNode)
            return True
        except (KeyError, TypeError):
            return False

    def _parent_exists(self, key: str) -> bool:
        """Check if the parent of the given path exists and is a GroupNode or RootNode.

        If the path itself is the root path, it returns False.

        Args:
            key (str): The path to check.

        Returns:
            bool: True if the parent exists and is a GroupNode or RootNode, False otherwise.
        """
        # Return false if the actual path is the root path
        if self._is_rootnode(key):
            return False

        parent, _ = split_path(key)
        try:
            self.nodes(parent, GroupNode, RootNode)
            return True
        except (KeyError, TypeError):
            return False

    def memory_bytes(self) -> int:
        """Returns the total memory size of the DataContainer in bytes.

        This includes the memory size of all nodes and their attributes.
        """
        return getsizeof(self) + getsizeof(self.nodes) + sum(n.memory_bytes() for n in self.nodes.values())

    def print_memory_usage(self):
        """Print detailed memory usage breakdown by path."""
        total_bytes = self.memory_bytes()

        print(f"DataContainer Memory Usage: {format_bytes(total_bytes)}")
        print("-" * len(f"DataContainer Memory Usage: {format_bytes(total_bytes)}"))

        # Collect node sizes
        node_sizes = []
        for path, node in self.nodes.items():
            size_bytes = node.memory_bytes()
            percentage = (size_bytes / total_bytes) * 100
            node_sizes.append((path, size_bytes, percentage, node.type.value))

        # Sort by size (largest first)
        node_sizes.sort(key=lambda x: x[1], reverse=True)

        # Print breakdown
        for path, size_bytes, percentage, node_type in node_sizes:
            size_str = format_bytes(size_bytes)
            print(f"{path:<30} {size_str:>10} ({percentage:>5.1f}%) [{node_type}]")



from typing import Protocol, runtime_checkable

from ...data import DataContainer
from ..utils import IOPathWrapper


@runtime_checkable
class BaseParser(Protocol):
    """A base class for all parsers that are used to parse data from a BytesIO object into a DataContainer object.

    All Subclasses must implement the parse method, which reads the data from the BytesIO object and returns it as a
    DataContainer object. Integrity checks and error handling should be implemented by the subclasses.
    """

    supported_extensions: tuple[str, ...] = ("",)

    def __init__(self) -> None:
        raise TypeError("This class is static and should not be instantiated.")

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        # Skip validation for ParserBase itself
        if cls.__name__ == "ParserBase":
            return

        # Check if supported_extensions is defined and not empty
        if (
            not hasattr(cls, "supported_extensions")
            or cls.supported_extensions == ("",)
            or not isinstance(cls.supported_extensions, tuple)
        ):
            raise TypeError(
                f"Parser class '{cls.__name__}' must define a classvariable called 'supported_extensions' "
                "as a tuple with at least one file extension in it"
            )

    @staticmethod
    def parse(data: IOPathWrapper) -> DataContainer:
        """Parses the data from the given IOPathWrapper object into a DataContainer object.

        Subclasses must implement this method.

        Args:
            data (IOPathWrapper): IOPathWrapper object containing the data to be parsed.

        Returns:
            DataContainer: The parsed data as a DataContainer object.
        """
        raise NotImplementedError("Subclasses must implement this method")



from collections.abc import Iterator, Sequence

from ..data import DataContainer
from ..transforms.base import _BaseTransform
from .base_dataset import BaseDataset


class IndexedThermoDataset(BaseDataset):
    """Extension of ThermoDataset that supports indexing with optional additional transforms.

    The IndexedThermoDataset maintains a subset of the parent dataset and allows for an additional transform to be
    applied to the data. This can be useful when a subset of the data needs to be selected and a different transform
    needs to be applied to the subset, e.g. for random splits of train, validation and test data. The
    IndexedThermoDataset maintains the transform chain of the parent dataset and appends the additional transform to it.
    """

    def __init__(self, dataset: BaseDataset, indices: Sequence[int], transform: _BaseTransform | None = None):
        """Initialize an indexed dataset with optional additional transform.

        Args:
            dataset (ThermoDataset): Parent dataset to index into
            indices (Sequence[int]): Sequence of indices to select from parent
            transform (_BaseTransform, optional): Optional transform to apply after parent's transform

        Raises:
            IndexError: If any of the provided indices are out of range
        """
        # Initialize base with dataset as parent and given transform
        super().__init__(parent=dataset, transform=transform)

        # Validate the indices
        if not all(0 <= i < len(dataset) for i in indices):
            raise IndexError(f"Provided indices are out of range. Must be within [0, {max(len(dataset) - 1, 0)}]")

        # Store parent dataset and indices
        self.__parent_dataset = dataset  # Parent dataset to avoid type issues with BaseDataset.parent
        self.__indices = indices  # Indices for subset

    def __len__(self) -> int:
        """Return length of indexed dataset."""
        return len(self.__indices)

    def __iter__(self) -> Iterator[DataContainer]:
        """Return iterator over indexed subset."""
        return (self[i] for i in range(len(self.__indices)))

    def load_raw_data(self, idx: int) -> DataContainer:
        """Load raw data from the parent dataset.

        Args:
            idx (int): Index into the subset

        Returns:
            DataContainer: Raw data container from the parent dataset
        """
        # Validate index
        if idx < 0 or idx >= len(self):
            msg = f"Index {idx} out of range."
            raise IndexError(msg + (f" Must be within [0, {len(self) - 1}]" if len(self) > 0 else " Empty dataset"))

        # Load raw data using parent's logic
        return self.__parent_dataset.load_raw_data(self.__indices[idx])

    def _map_index(self, idx: int) -> int:
        """Map the index to the parent's index."""
        return self.__indices[idx]

    @property
    def files(self) -> list[str]:
        """Return list of files corresponding to indexed subset."""
        return [self.__parent_dataset.files[i] for i in self.__indices]



from .azure_backend import AzureBlobBackend
from .base_backend import BaseBackend
from .local_backend import LocalBackend
from .s3_backend import S3Backend

__all__ = ["AzureBlobBackend", "BaseBackend", "LocalBackend", "S3Backend"]



import logging

from .__pkginfo__ import __version__
from .config import configure_logging, settings
from .data import DataContainer, ThermoContainer
from .dataset import IndexedThermoDataset, ThermoDataset
from .io import HDF5Parser, SimulationParser
from .readers import LocalReader, S3Reader
from .transforms import augmentation, normalization, preprocessing, sampling, utils
from .writers import LocalWriter, S3Writer

# Set up logging per Python best practices: https://docs.python.org/3/howto/logging.html
# Add NullHandler to prevent "No handlers could be found" warnings
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "DataContainer",
    "HDF5Parser",
    "IndexedThermoDataset",
    "LocalReader",
    "LocalWriter",
    "S3Reader",
    "S3Writer",
    "SimulationParser",
    "ThermoContainer",
    "ThermoDataset",
    "__version__",
    "augmentation",
    "configure_logging",
    "normalization",
    "preprocessing",
    "sampling",
    "settings",
    "utils",
]

# TODO: Implement async data loading



import json
from typing import Any

import numpy as np
import pymatreader
from scipy.io.matlab import MatReadError

from ...data import DataContainer, ThermoContainer
from ..utils import IOPathWrapper
from .base_parser import BaseParser


class SimulationParser(BaseParser):
    supported_extensions = (".mat",)

    @staticmethod
    def parse(data: IOPathWrapper) -> DataContainer:  # pylint: disable=too-many-branches
        """Parses the data from the given IOPathWrapper object into a DataContainer object.

        The IOPathWrapper object must contain a .mat file with simulattion data from COMSOL.

        Args:
            data (IOPathWrapper): IOPathWrapper object containing the data to be parsed.

        Returns:
            DataContainer: The parsed data as a DataContainer object.

        Raises:
            ValueError: If the given IOPathWrapper object is empty or does not contain a valid .mat file.
        """
        # Check if the IOPathWrapper object is empty
        if data.file_obj.getbuffer().nbytes == 0:
            raise ValueError("The given IOPathWrapper object is empty.")

        # Try to load the .mat file using pymatreader ==> If it fails the file is not a valid .mat file
        try:
            data_dict = pymatreader.read_mat(data.file_path)["SimResult"]
        except MatReadError as o:
            raise ValueError("The given IOPathWrapper object does not contain a valid .mat file.") from o

        # Create an empty Thermocontainer ==> predefined structure
        datacontainer = ThermoContainer()

        # Add source as an attribute
        datacontainer.add_attributes(path="/MetaData", Source="Simulation")

        # Iterate through all keys and save the values in the datacontainer ==>
        # If one key does not exist the variable in the datapoint will stay None
        for key in data_dict.keys():
            match key:
                case "Tdata":
                    datacontainer.update_dataset(path="/Data/Tdata", data=data_dict[key])
                case "GroundTruth":
                    # Try to extract the label ids and convert to dict
                    try:
                        label_ids = data_dict[key]["LabelIds"]
                        label_ids = json.loads(label_ids) if isinstance(label_ids, str) else label_ids
                        datacontainer.add_attributes(path="/GroundTruth/DefectMask", LabelIds=label_ids)
                    except (json.JSONDecodeError, AttributeError, KeyError):
                        pass
                    datacontainer.update_dataset(path="/GroundTruth/DefectMask", data=data_dict[key]["DefectMask"])
                case "Time":
                    datacontainer.update_dataset(path="/MetaData/DomainValues", data=data_dict[key])
                case "LookUpTable":
                    datacontainer.update_dataset(path="/MetaData/LookUpTable", data=data_dict[key])
                case "ExcitationSignal":
                    datacontainer.update_dataset(path="/MetaData/ExcitationSignal", data=data_dict[key])
                case "ComsolParameters":
                    # Reshape comsol parameters back to a list of lists ==> pymatreader loads this as a flattened list
                    comsol_parameters = reshape_pymatreader_parameters(data_dict[key])

                    # Clean up the list of lists ==> handle empty arrays and single dimensions
                    for i, sublist in enumerate(comsol_parameters):
                        for j, item in enumerate(sublist):
                            if isinstance(item, np.ndarray) and item.size == 1:
                                comsol_parameters[i][j] = item.item()
                            if isinstance(item, np.ndarray) and item.size == 0:
                                comsol_parameters[i][j] = ""

                    # TODO: Actually this is a list of lists. Should be improved in the future
                    # (maybe with a pandas dataframe ==> needs more work!)
                    # Replace ' with " to make it a valid json string
                    converted_comsol_parameters = str(comsol_parameters).replace("'", '"')

                    # Replace nan with NaN to make it a valid json string
                    converted_comsol_parameters = converted_comsol_parameters.replace("nan", "NaN")

                    # Try to load the json string into a python list
                    try:
                        sim_par = json.loads(converted_comsol_parameters)
                    # If it fails just save the raw string
                    except json.JSONDecodeError:
                        sim_par = converted_comsol_parameters
                    datacontainer.add_attributes(path="/MetaData", SimulationParameter=sim_par)

                case "NoiseLevel":
                    datacontainer.add_attributes(path="/Data/Tdata", NoiseLevel=data_dict[key])
                case "Shapes":
                    # Convert the Shapes into a Python dict first:
                    shapes = {
                        data_dict[key]["Names"][i]: int(data_dict[key]["Numbers"][i])
                        for i in range(len(data_dict[key]["Names"]))
                    }
                    datacontainer.add_attributes(path="/MetaData", Shapes=shapes)

        # Return the constructed datapoint
        return datacontainer


def reshape_pymatreader_parameters(flat_params: list[Any]) -> list[list[Any]]:
    """Reshape the flattened list of parameters from pymatreader into a list of lists."""
    # Determine where the sections start
    num_params = len(flat_params) // 5

    # Extract each section
    param_names = flat_params[:num_params]
    expressions = flat_params[num_params : 2 * num_params]
    descriptions = flat_params[2 * num_params : 3 * num_params]
    values = flat_params[3 * num_params : 4 * num_params]
    units = flat_params[4 * num_params : 5 * num_params]

    # Create the column headers
    result = []

    # Create a row for each parameter
    for i in range(num_params):
        result.append([param_names[i], expressions[i], descriptions[i], values[i], units[i]])

    return result



from tqdm.auto import tqdm

TQDM_DEFAULT_KWARGS = {
    "unit": "B",
    "unit_scale": True,
    "unit_divisor": 1024,
    "leave": False,
}


def get_tqdm_delay(file_size: int, delay: float = 4.0) -> float:
    """Get delay value based on file size."""
    return delay if file_size < 40 * 1024 * 1024 else 0


def get_tqdm_default_kwargs(file_size: int = 0, delay: float = 4.0) -> dict:
    """Get default tqdm kwargs."""
    return {**TQDM_DEFAULT_KWARGS, "total": file_size, "delay": get_tqdm_delay(file_size, delay)}


class TqdmCallback(tqdm):
    """Progress bar with callback support.

    This allows to display a progress bar for long running operations that provide a callback function.
    """

    def __init__(self, total: int, desc: str, delay: float = 4.0, **kwargs):
        """Initialize with automatic delay-based display.

        Args:
            total (int): Total bytes expected
            desc (str): Progress description
            delay (float, optional): Show progress only if operation takes > delay seconds. Only for files < 40MB
            **kwargs: Additional tqdm arguments to be passed to tqdm
        """
        merged_kwargs = {
            **TQDM_DEFAULT_KWARGS,
            "total": total,
            "desc": desc,
            "delay": get_tqdm_delay(total, delay),
            **kwargs,  # User overrides
        }
        super().__init__(**merged_kwargs)

    def callback(self, bytes_amount: int) -> None:
        """Callback function for boto3 operations."""
        self.update(bytes_amount)



from azure.core.credentials import TokenCredential

from ..io import AzureBlobBackend, BaseParser
from .base_reader import BaseReader


class AzureBlobReader(BaseReader):
    def __init__(
        self,
        account_url: str,
        container_name: str,
        prefix: str,
        connection_string: str | None = None,
        credential: TokenCredential | None = None,
        num_files: int | None = None,
        download_files: bool = False,
        cache_files: bool = True,
        parser: type[BaseParser] | None = None,
    ):
        """Initialize reader for Azure Blob Storage.

        Uses DefaultAzureCredential by default (managed identity, az login, env vars, etc.).
        Explicit connection_string or credential override the default behavior.

        Args:
            account_url (str): Storage account URL (https://<account>.blob.core.windows.net)
            container_name (str): Azure blob container name
            prefix (str): Prefix (folder path) within container
            connection_string (str | None): Optional connection string (overrides default auth)
            credential (TokenCredential | None): Optional Azure TokenCredential (overrides default auth)
            num_files (int | None): Maximum number of files to read. If None, reads all files.
            download_files (bool): If True, downloads and caches files locally during operations.
                If False, files are accessed on-demand without local caching. Default: False.
            cache_files (bool): If True, caches the file list in memory. If False, file list is
                refreshed on each access. Default: True.
            parser (type[BaseParser] | None): Parser class for reading files. If None, auto-selects
                based on file extension. Default: None.
        """
        super().__init__(num_files, download_files, cache_files, parser)
        self.__container_name = container_name
        self.__prefix = prefix
        self.__connection_string = connection_string
        self.__account_url = account_url
        self.__credential = credential

    def _create_backend(self) -> AzureBlobBackend:
        """Create a new AzureBlobBackend instance.

        Called to create or recreate the backend when needed or after unpickling.
        """
        return AzureBlobBackend(
            account_url=self.__account_url,
            container_name=self.__container_name,
            prefix=self.__prefix,
            connection_string=self.__connection_string,
            credential=self.__credential,
        )

    def _get_reader_params(self) -> str:
        """Get string representation of reader parameters."""
        auth_method = (
            "connection_string"
            if self.__connection_string
            else "credential"
            if self.__credential
            else "DefaultAzureCredential"
        )
        return f"container={self.__container_name}, prefix={self.__prefix}, auth={auth_method}"



import sys
from dataclasses import asdict, dataclass


@dataclass(frozen=True)
class Unit:
    """Dataclass that stores unit information."""

    name: str  # The name of the unit (e.g. kelvin, celsius, etc.)
    quantity: str  # The quantity the unit represents (e.g. temperature, time, etc.)
    symbol: str  # The symbol of the unit (e.g. K, °C, etc.)

    def to_dict(self) -> dict:
        """Convert the Unit dataclass to a dictionary.

        Returns:
            dict: The dictionary representation of the Unit.
        """
        return asdict(self)


def print_available_units() -> None:
    """Print all available units."""
    module = sys.modules[__name__]
    for name in dir(module):
        if isinstance(obj := getattr(module, name), Unit):
            print(obj)


# Units definitions
# Special units
dimensionless = Unit(name="dimensionless", quantity="dimensionless", symbol="1")
# arbitrary: Special unit for data that is not really dimensionless but has not been processed yet
# (e.g. temp without LUT applied)
arbitrary = Unit(name="arbitrary", quantity="arbitrary", symbol="a. u.")
# undefined: Return value for datasets without a unit defined
undefined = Unit(name="undefined", quantity="undefined", symbol="N/A")

# Temperature
kelvin = Unit(name="kelvin", quantity="temperature", symbol="K")
celsius = Unit(name="celsius", quantity="temperature", symbol="°C")

# Time
second = Unit(name="second", quantity="time", symbol="s")
millisecond = Unit(name="millisecond", quantity="time", symbol="ms")

# Frequency
hertz = Unit(name="hertz", quantity="frequency", symbol="Hz")

# Add more units here as needed



import warnings

from ..data import DataContainer
from ..data.units import dimensionless
from .base import ThermoTransform


class MinMaxNormalize(ThermoTransform):
    """Normalizes the Temperature data (Tdata) in the container to range [0, 1].

    This is done by subtracting the min value from the data and dividing by the difference between the max and
    min values.
    """

    def __init__(self, eps: float = 1e-12):
        """Normalizes the Temperature data (Tdata) in the container to range [0, 1].

        This is done by subtracting the min value from the data and dividing by the difference between the max and
        min values.

        Args:
            eps (float): Small value added to the denominator to avoid division by zero. Default is 1e-12.
        """
        super().__init__()
        if eps > 1e-3:
            warnings.warn(
                f"eps={eps} is larger than 1e-3. This may significantly affect "
                "normalization results beyond numerical stability.",
                stacklevel=2,
            )
        self.eps = eps

    def forward(self, container: DataContainer) -> DataContainer:
        # Extract the data
        tdata = container.get_dataset("/Data/Tdata")

        # Get min and max values of the tensor and normalize the data
        min_val = tdata.min()
        max_val = tdata.max()
        tdata = (tdata - min_val) / (max_val - min_val + self.eps)

        # Update the container and return it
        container.update_dataset("/Data/Tdata", tdata)
        container.update_unit("/Data/Tdata", dimensionless)
        return container


class MaxNormalize(ThermoTransform):
    """Normalizes the Temperature data (Tdata) in the container to range [0, 1] by using the max value of the data.

    This is done by dividing the data by the max value of the data.
    """

    def __init__(self, eps: float = 1e-12):
        """Normalizes the Temperature data (Tdata) in the container to range [0, 1], by using the max value of the data.

        Args:
            eps (float): Small value added to the denominator to avoid division by zero. Default is 1e-12.
        """
        super().__init__()
        if eps > 1e-3:
            warnings.warn(
                f"eps={eps} is larger than 1e-3. This may significantly affect "
                "normalization results beyond numerical stability.",
                stacklevel=2,
            )
        self.eps = eps

    def forward(self, container: DataContainer) -> DataContainer:
        # Extract the data
        tdata = container.get_dataset("/Data/Tdata")

        # Get max value of the tensor and normalize the data
        max_val = tdata.max()
        tdata = tdata / (max_val + self.eps)

        # Update the container and return it
        container.update_dataset("/Data/Tdata", tdata)
        container.update_unit("/Data/Tdata", dimensionless)
        return container


class ZScoreNormalize(ThermoTransform):
    """Normalizes the Temperature data (Tdata) in the container to have mean 0 and standard deviation 1.

    This is done by subtracting the mean value from the data and dividing by the standard deviation of the data.
    """

    def __init__(self, eps: float = 1e-12):
        """Normalizes the Temperature data (Tdata) in the container to have mean 0 and standard deviation 1.

        Args:
            eps (float): Small value added to the denominator to avoid division by zero. Default is 1e-12.
        """
        super().__init__()
        if eps > 1e-3:
            warnings.warn(
                f"eps={eps} is larger than 1e-3. This may significantly affect "
                "normalization results beyond numerical stability.",
                stacklevel=2,
            )
        self.eps = eps

    def forward(self, container: DataContainer) -> DataContainer:
        # Extract the data
        tdata = container.get_dataset("/Data/Tdata")

        # Get mean and standard deviation of the tensor and normalize the data
        mean_val = tdata.mean()
        std_val = tdata.std()
        tdata = (tdata - mean_val) / (std_val + self.eps)

        # Update the container and return it
        container.update_dataset("/Data/Tdata", tdata)
        container.update_unit("/Data/Tdata", dimensionless)
        return container
